##
# .test.test_notifyman - test .notifyman
##
import unittest
import threading
import time
from ..temporal import pg_tmp
from ..notifyman import NotificationManager

class test_notifyman(unittest.TestCase):
	@pg_tmp
	def testNotificationManager(self):
		# signals each other
		alt = new()
		with alt:
			nm = NotificationManager(db, alt)
			db.listen('foo')
			alt.listen('bar')
			# notify the other.
			alt.notify('foo')
			db.notify('bar')
			# we can separate these here because there's no timeout
			for ndb, notifies in nm:
				for n in notifies:
					if ndb is db:
						self.assertEqual(n[0], 'foo')
						self.assertEqual(n[1], '')
						self.assertEqual(n[2], alt.backend_id)
						nm.connections.discard(db)
					elif ndb is alt:
						self.assertEqual(n[0], 'bar')
						self.assertEqual(n[1], '')
						self.assertEqual(n[2], db.backend_id)
						nm.connections.discard(alt)
					else:
						self.fail("unknown connection received notify..")

	@pg_tmp
	def testNotificationManagerTimeout(self):
		nm = NotificationManager(db, timeout = 0.1)
		db.listen('foo')
		count = 0
		for event in nm:
			if event is None:
				# do this a few times, then break out of the loop
				db.notify('foo')
				continue
			ndb, notifies = event
			self.assertEqual(ndb, db)
			for n in notifies:
				self.assertEqual(n[0], 'foo')
				self.assertEqual(n[1], '')
				self.assertEqual(n[2], db.backend_id)
				count = count + 1
			if count > 3:
				break

	@pg_tmp
	def testNotificationManagerZeroTimeout(self):
		# Zero-timeout means raise StopIteration when
		# there are no notifications to emit.
		# It checks the wire, but does *not* wait for data.
		nm = NotificationManager(db, timeout = 0)
		db.listen('foo')
		self.assertEqual(list(nm), [])
		db.notify('foo')
		time.sleep(0.01)
		self.assertEqual(list(nm), [('foo','',db.backend_id)]) # bit of a race

	@pg_tmp
	def test_iternotifies(self):
		# db.iternotifies() simplification of NotificationManager
		alt = new()
		alt.listen('foo')
		alt.listen('close')
		def get_notices(db, l):
			with db:
				for x in db.iternotifies():
					if x[0] == 'close':
						break
					l.append(x)
		rl = []
		t = threading.Thread(target = get_notices, args = (alt, rl,))
		t.start()
		db.notify('foo')
		while not rl:
			time.sleep(0.05)
		channel, payload, pid = rl.pop(0)
		self.assertEqual(channel, 'foo')
		self.assertEqual(payload, '')
		self.assertEqual(pid, db.backend_id)
		db.notify('close')

	@pg_tmp
	def testNotificationManagerZeroTimeout(self):
		# Zero-timeout means raise StopIteration when
		# there are no notifications to emit.
		# It checks the wire, but does *not* wait for data.
		db.listen('foo')
		self.assertEqual(list(db.iternotifies(0)), [])
		db.notify('foo')
		time.sleep(0.01)
		self.assertEqual(list(db.iternotifies(0)), [('foo','', db.backend_id)]) # bit of a race

	@pg_tmp
	def testNotificationManagerOnClosed(self):
		# When the connection goes away, the NM iterator
		# should raise a Stop.
		db = new()
		db.listen('foo')
		db.notify('foo')
		for n in db.iternotifies():
			db.close()
		self.assertEqual(db.closed, True)
		del db
		# closer, after an idle
		db = new()
		db.listen('foo')
		for n in db.iternotifies(0.2):
			if n is None:
				# In the loop, notify, and expect to
				# get the notification even though the
				# connection was closed.
				db.notify('foo')
				db.execute('')
				db.close()
				hit = False
			else:
				hit = True
		# hit should get set two times.
		# once on the first idle, and once on the event
		# received after the close.
		self.assertEqual(db.closed, True)
		self.assertEqual(hit, True)

if __name__ == '__main__':
	unittest.main()



##
# .protocol.buffer
##
"""
This is an abstraction module that provides the working buffer implementation.
If a C compiler is not available on the system that built the package, the slower
`postgresql.protocol.pbuffer` module can be used in
`postgresql.port.optimized.buffer`'s absence.

This provides a convenient place to import the necessary module without
concerning the local code with the details.
"""
try:
	from ..port.optimized import pq_message_stream
except ImportError:
	from .pbuffer import pq_message_stream



##
# .configfile
##
"""
PostgreSQL configuration file parser and editor functions.
"""
import sys
import os
from . import string as pg_str
from . import api as pg_api

quote = "'"
comment = '#'

def parse_line(line, equality = '=', comment = comment, quote = quote):
	keyval = line.split(equality, 1)
	if len(keyval) == 2:
		key, val = keyval

		prekey_len = 0
		for c in key:
			if not c.isspace() and c not in comment:
				break
			prekey_len += 1

		key_len = 0
		for c in key[prekey_len:]:
			if not (c.isalpha() or c.isdigit() or c in '_'):
				break
			key_len += 1

		# If non-whitespace exists after the key,
		# it's a complex comment, so just bail out.
		if key[prekey_len + key_len:].strip():
			return

		preval_len = 0
		for c in val:
			if not c.isspace() or c in '\n\r':
				break
			preval_len += 1

		inquotes = False
		escaped = False
		val_len = 0
		for i in range(preval_len, len(val)):
			c = val[i]
			if c == quote:
				if inquotes is False:
					inquotes = True
				else:
					if escaped is False:
						# Peek ahead to see if it's escaped with another quote
						escaped = (len(val) > i+1 and val[i+1] == quote)
						if escaped is False:
							inquotes = False
					elif escaped is True:
						# It *was* an escaped quote.
						escaped = False
			elif inquotes is False and (c.isspace() or c in comment):
				break
			val_len += 1

		return (
			# The key slice
			slice(prekey_len, key_len + prekey_len),
			# The value slice
			slice(len(key) + 1 + preval_len, len(key) + 1 + preval_len + val_len)
		)

def unquote(s, quote = quote):
	"""
	Unquote the string `s` if quoted.
	"""
	s = s.strip()
	if not s.startswith(quote):
		return s
	return s[1:-1].replace(quote*2, quote)

def write_config(map, writer, keys = None):
	"""
	A configuration writer that will trample & merely write the settings.
	"""
	if keys is None:
		keys = map
	for k in keys:
		writer('='.join((k, map[k])) + os.linesep)

def alter_config(map, fo, keys = None):
	"""
	Alters a configuration file without trampling on the existing structure.
	"""
	if keys is None:
		keys = list(map.keys())
	# Normalize keys and map them back to
	pkeys = {
		k.lower().strip() : keys.index(k) for k in keys
	}

	lines = []
	candidates = {}
	i = -1
	# Process lines in fo
	for l in fo:
		i += 1
		lines.append(l)
		pl = parse_line(l)
		# "Bad" line? fuh-get-duh-bowt-it.
		if pl is None:
			continue
		sk, sv = pl
		k = l[sk].lower()
		v = l[sv]
		# It's a candidate?
		if k in pkeys:
			c = candidates.get(k)
			if c is None:
				candidates[k] = c = []
			c.append((i, sk, sv))
	# Simply insert the data somewhere for unfound keys.
	for k in pkeys:
		if k not in candidates:
			key = keys[pkeys[k]]
			val = map[key]
			# Can't comment without an uncommented candidate.
			if val is not None:
				if not lines[-1].endswith(os.linesep):
					lines[-1] = lines[-1] + os.linesep
				lines.append("%s = '%s'" %(key, val.replace("'", "''")))

	# Multiple lines may have the key, so make a decision based on the value.
	for ck in candidates.keys():
		to_set_key = keys[pkeys[ck]]
		to_set_val = map[keys[pkeys[ck]]]

		if to_set_val is None:
			# Comment uncommented occurrences.
			for cl in candidates[ck]:
				line_num, sk, sv = cl
				if comment not in lines[line_num][:sk.start]:
					lines[line_num] = '#' + lines[line_num]
		else:
			# Manage occurrences.
			# w_ is for winner.
			# Now, a winner is elected for alteration. The winner
			# is decided based on a two factors: commenting and value.
			w_score = -1
			w_commented = None
			w_val = None
			w_cl = None
			for cl in candidates[ck]:
				line_num, sk, sv = cl
				l = lines[line_num]
				lkey = l[sk]
				lval = l[sv]
				commented = (comment in l[:sk.start])
				score = \
					(not commented and 1 or 0) + \
					(unquote(lval) == to_set_val and 2 or 0)
				# So, if a line is not commented, and has equal
				# values, then that's the winner. If a line is commented,
				# and has a has equal values, it will succeed over a mere
				# uncommented value.

				if score > w_score:
					if w_commented is False:
						# It's now a loser, so comment it out if necessary.
						lines[w_cl[0]] = '#' + lines[w_cl[0]]
					w_score = score
					w_commented = commented
					w_val = lval
					w_cl = cl
				elif commented is False:
					# Loser needs to be commented.
					lines[line_num] = '#' + l

			line_num, sk, sv = w_cl
			l = lines[line_num]
			if w_commented:
				bol = ''
			else:
				bol = l[:sk.start]
			post_val = l[sv.stop:]
			# If there is post-value data, validate that it's commented.
			if post_val and not post_val.isspace():
				stripped_post_val = post_val.lstrip()
				if not stripped_post_val.startswith(comment):
					post_val = '%s%s%s' %(
						# The whitespace before the uncommented visibles
						post_val[0:len(post_val) - len(stripped_post_val)],
						# A comment followed by the uncommented visibles
						comment, stripped_post_val
					)
			# Everything is set as quoted as it's the only safe
			# way to set something without delving into setting types.
			lines[line_num] = \
				bol + l[sk.start:sv.start] + \
				"'%s'" %(to_set_val.replace("'", "''"),) + post_val
	return lines

def read_config(iter, d = None, selector = None):
	if d is None:
		d = {}
	for line in iter:
		kv = parse_line(line)
		if kv:
			key = line[kv[0]]
			if comment not in line[:kv[0].start] and \
			(selector is None or selector(key)):
				d[key] = unquote(line[kv[1]])
	return d

class ConfigFile(pg_api.Settings):
	"""
	Provides a mapping interface to a configuration file.

	Every operation will cause the file to be wholly read, so using `update` to make
	multiple changes is desirable.
	"""
	_e_factors = ('path',)
	_e_label = 'CONFIGFILE'

	def _e_metas(self):
		yield (None, len(self.keys()))

	def __init__(self, path, open = open):
		self.path = path
		self._open = open
		self._store = []
		self._restore = {}

	def __repr__(self):
		return "%s.%s(%r)" %(
			type(self).__module__,
			type(self).__name__,
			self.path
		)

	def _save(self, lines : [str]):
		with self._open(self.path, 'w') as cf:
			for l in lines:
				cf.write(l)

	def __delitem__(self, k):
		with self._open(self.path) as cf:
			lines = alter_config({k : None}, cf)
		self._save()

	def __getitem__(self, k):
		with self._open(self.path) as cfo:
			return read_config(
				cfo,
				selector = k.__eq__
			)[k]

	def __setitem__(self, k, v):
		self.update({k : v})

	def __call__(self, **kw):
		self._store.insert(0, kw)

	def __context__(self):
		return self

	def __iter__(self):
		return self.keys()

	def __len__(self):
		return len(list(self.keys()))

	def __enter__(self):
		res = self.getset(self._store[0].keys())
		self.update(self._store[0])
		del self._store[0]
		self._restore.append(res)

	def __exit__(self, exc, val, tb):
		self._restored.update(self._restore[-1])
		del self._restore[-1]
		self.update(self._restored)
		self._restored.clear()
		return exc is None

	def get(self, k, alt = None):
		with self._open(self.path) as cf:
			return read_config(cf, selector = k.__eq__).get(k, alt)

	def keys(self):
		return read_config(self._open(self.path)).keys()

	def values(self):
		return read_config(self._open(self.path)).values()

	def items(self):
		return read_config(self._open(self.path)).items()

	def update(self, keyvals):
		"""
		Given a dictionary of settings, apply them to the cluster's
		postgresql.conf.
		"""
		with self._open(self.path) as cf:
			lines = alter_config(keyvals, cf)
		self._save(lines)

	def getset(self, keys):
		"""
		Get all the settings in the list of keys.
		Returns a dictionary of those keys.
		"""
		keys = set(keys)
		with self._open(self.path) as cfo:
			cfg = read_config(
				cfo,
				selector = keys.__contains__
			)
			for x in (keys - set(cfg.keys())):
				cfg[x] = None
			return cfg



##
# .python.itertools
##
"""
itertools extensions
"""
import collections.abc
from itertools import cycle, islice

def interlace(*iters, next = next) -> collections.abc.Iterable:
	"""
	interlace(i1, i2, ..., in) -> (
		i1-0, i2-0, ..., in-0,
		i1-1, i2-1, ..., in-1,
		.
		.
		.
		i1-n, i2-n, ..., in-n,
	)
	"""
	return map(next, cycle([iter(x) for x in iters]))

def chunk(iterable, chunksize = 256):
	"""
	Given an iterable, return an iterable producing chunks of the objects
	produced by the given iterable.

	chunks([o1,o2,o3,o4], chunksize = 2) -> [
		[o1,o2],
		[o3,o4],
	]
	"""
	iterable = iter(iterable)
	last = ()
	lastsize = chunksize
	while lastsize == chunksize:
		last = list(islice(iterable, chunksize))
		lastsize = len(last)
		yield last

def find(iterable, selector):
	"""
	Return the first item in the `iterable` that causes the `selector` to return
	`True`.
	"""
	for x in iterable:
		if selector(x):
			return x



##
# .test.test_protocol
##
import sys
import unittest
import struct
import decimal
import socket
import time
from threading import Thread

from ..protocol import element3 as e3
from ..protocol import xact3 as x3
from ..protocol import client3 as c3
from ..protocol import buffer as pq_buf
from ..python.socket import find_available_port, SocketFactory

def pair(msg):
	return (msg.type, msg.serialize())
def pairs(*msgseq):
	return list(map(pair, msgseq))

long = struct.Struct("!L")
packl = long.pack
unpackl = long.unpack

class test_buffer(unittest.TestCase):
	def setUp(self):
		self.buffer = pq_buf.pq_message_stream()

	def testMultiByteMessage(self):
		b = self.buffer
		b.write(b's')
		self.assertTrue(b.next_message() is None)
		b.write(b'\x00\x00')
		self.assertTrue(b.next_message() is None)
		b.write(b'\x00\x10')
		self.assertTrue(b.next_message() is None)
		data = b'twelve_chars'
		b.write(data)
		self.assertEqual(b.next_message(), (b's', data))

	def testSingleByteMessage(self):
		b = self.buffer
		b.write(b's')
		self.assertTrue(b.next_message() is None)
		b.write(b'\x00')
		self.assertTrue(b.next_message() is None)
		b.write(b'\x00\x00\x05')
		self.assertTrue(b.next_message() is None)
		b.write(b'b')
		self.assertEqual(b.next_message(), (b's', b'b'))

	def testEmptyMessage(self):
		b = self.buffer
		b.write(b'x')
		self.assertTrue(b.next_message() is None)
		b.write(b'\x00\x00\x00')
		self.assertTrue(b.next_message() is None)
		b.write(b'\x04')
		self.assertEqual(b.next_message(), (b'x', b''))

	def testInvalidLength(self):
		b = self.buffer
		b.write(b'y\x00\x00\x00\x03')
		self.assertRaises(ValueError, b.next_message,)

	def testRemainder(self):
		b = self.buffer
		b.write(b'r\x00\x00\x00\x05Aremainder')
		self.assertEqual(b.next_message(), (b'r', b'A'))

	def testLarge(self):
		b = self.buffer
		factor = 1024
		r = 10000
		b.write(b'X' + packl(factor * r + 4))
		segment = b'\x00' * factor
		for x in range(r-1):
			b.write(segment)
		b.write(segment)
		msg = b.next_message()
		self.assertTrue(msg is not None)
		self.assertEqual(msg[0], b'X')

	def test_getvalue(self):
		# Make sure that getvalue() only applies to messages
		# that have not been read.
		b = self.buffer
		# It should be empty.
		self.assertEqual(b.getvalue(), b'')
		d = b'F' + packl(28)
		b.write(d)
		self.assertEqual(b.getvalue(), d)
		d1 = b'01'*12 # 24
		b.write(d1)
		self.assertEqual(b.getvalue(), d + d1)
		out = b.read()[0]
		self.assertEqual(out, (b'F', d1))
		nd = b'N'
		b.write(nd)
		self.assertEqual(b.getvalue(), nd)
		b.write(packl(4))
		self.assertEqual(list(b.read()), [(b'N', b'')])
		self.assertEqual(b.getvalue(), b'')
		# partial; read one message to exercise
		# that the appropriate fragment of the first
		# chunk in the buffer is picked up.
		first_body = (b'1234' * 3)
		first = b'v' + packl(len(first_body) + 4) + first_body
		second_body = (b'4321' * 5)
		second = b'z' + packl(len(second_body) + 4) + second_body
		b.write(first + second)
		self.assertEqual(b.getvalue(), first + second)
		self.assertEqual(list(b.read(1)), [(b'v', first_body)])
		self.assertEqual(b.getvalue(), second)
		self.assertEqual(list(b.read(1)), [(b'z', second_body)])
		# now, with a third full message in the next chunk
		third_body = (b'9876' * 10)
		third = b'3' + packl(len(third_body) + 4) + third_body
		b.write(first + second)
		b.write(third)
		self.assertEqual(b.getvalue(), first + second + third)
		self.assertEqual(list(b.read(1)), [(b'v', first_body)])
		self.assertEqual(b.getvalue(), second + third)
		self.assertEqual(list(b.read(1)), [(b'z', second_body)])
		self.assertEqual(b.getvalue(), third)
		self.assertEqual(list(b.read(1)), [(b'3', third_body)])
		self.assertEqual(b.getvalue(), b'')

##
# element3 tests
##

message_samples = [
	e3.VoidMessage,
	e3.Startup([
		(b'user', b'jwp'),
		(b'database', b'template1'),
		(b'options', b'-f'),
	]),
	e3.Notice((
		(b'S', b'FATAL'),
		(b'M', b'a descriptive message'),
		(b'C', b'FIVEC'),
		(b'D', b'bleh'),
		(b'H', b'dont spit into the fan'),
	)),
	e3.Notify(123, b'wood_table'),
	e3.KillInformation(19320, 589483),
	e3.ShowOption(b'foo', b'bar'),
	e3.Authentication(4, b'salt'),
	e3.Complete(b'SELECT'),
	e3.Ready(b'I'),
	e3.CancelRequest(4123, 14252),
	e3.NegotiateSSL(),
	e3.Password(b'ckr4t'),
	e3.AttributeTypes(()),
	e3.AttributeTypes(
		(123,) * 1
	),
	e3.AttributeTypes(
		(123,0) * 1
	),
	e3.AttributeTypes(
		(123,0) * 2
	),
	e3.AttributeTypes(
		(123,0) * 4
	),
	e3.TupleDescriptor(()),
	e3.TupleDescriptor((
		(b'name', 123, 1, 1, 0, 0, 1,),
	)),
	e3.TupleDescriptor((
		(b'name', 123, 1, 2, 0, 0, 1,),
	) * 2),
	e3.TupleDescriptor((
		(b'name', 123, 1, 2, 1, 0, 1,),
	) * 3),
	e3.TupleDescriptor((
		(b'name', 123, 1, 1, 0, 0, 1,),
	) * 1000),
	e3.Tuple([]),
	e3.Tuple([b'foo',]),
	e3.Tuple([None]),
	e3.Tuple([b'foo',b'bar']),
	e3.Tuple([None, None]),
	e3.Tuple([None, b'foo', None]),
	e3.Tuple([b'bar', None, b'foo', None, b'bleh']),
	e3.Tuple([b'foo', b'bar'] * 100),
	e3.Tuple([None] * 100),
	e3.Query(b'select * from u'),
	e3.Parse(b'statement_id', b'query', (123, 0)),
	e3.Parse(b'statement_id', b'query', (123,)),
	e3.Parse(b'statement_id', b'query', ()),
	e3.Bind(b'portal_id', b'statement_id',
		(b'tt',b'\x00\x00'),
		[b'data',None], (b'ff',b'xx')),
	e3.Bind(b'portal_id', b'statement_id', (b'tt',), [None], (b'xx',)),
	e3.Bind(b'portal_id', b'statement_id', (b'ff',), [b'data'], ()),
	e3.Bind(b'portal_id', b'statement_id', (), [], (b'xx',)),
	e3.Bind(b'portal_id', b'statement_id', (), [], ()),
	e3.Execute(b'portal_id', 500),
	e3.Execute(b'portal_id', 0),
	e3.DescribeStatement(b'statement_id'),
	e3.DescribePortal(b'portal_id'),
	e3.CloseStatement(b'statement_id'),
	e3.ClosePortal(b'portal_id'),
	e3.Function(123, (), [], b'xx'),
	e3.Function(321, (b'tt',), [b'foo'], b'xx'),
	e3.Function(321, (b'tt',), [None], b'xx'),
	e3.Function(321, (b'aa', b'aa'), [None,b'a' * 200], b'xx'),
	e3.FunctionResult(b''),
	e3.FunctionResult(b'foobar'),
	e3.FunctionResult(None),
	e3.CopyToBegin(123, [321,123]),
	e3.CopyToBegin(0, [10,]),
	e3.CopyToBegin(123, []),
	e3.CopyFromBegin(123, [321,123]),
	e3.CopyFromBegin(0, [10]),
	e3.CopyFromBegin(123, []),
	e3.CopyData(b''),
	e3.CopyData(b'foo'),
	e3.CopyData(b'a' * 2048),
	e3.CopyFail(b''),
	e3.CopyFail(b'iiieeeeee!'),
]

class test_element3(unittest.TestCase):
	def test_cat_messages(self):
		# The optimized implementation will identify adjacent copy data, and
		# take a more efficient route; so rigorously test the switch between the
		# two modes.
		self.assertEqual(e3.cat_messages([]), b'')
		self.assertEqual(e3.cat_messages([b'foo']), b'd\x00\x00\x00\x07foo')
		self.assertEqual(e3.cat_messages([b'foo', b'foo']), 2*b'd\x00\x00\x00\x07foo')
		# copy, other, copy
		self.assertEqual(e3.cat_messages([b'foo', e3.SynchronizeMessage, b'foo']),
			b'd\x00\x00\x00\x07foo' + e3.SynchronizeMessage.bytes() + b'd\x00\x00\x00\x07foo')
		# copy, other, copy*1000
		self.assertEqual(e3.cat_messages(1000*[b'foo', e3.SynchronizeMessage, b'foo']),
			1000*(b'd\x00\x00\x00\x07foo' + e3.SynchronizeMessage.bytes() + b'd\x00\x00\x00\x07foo'))
		# other, copy, copy*1000
		self.assertEqual(e3.cat_messages(1000*[e3.SynchronizeMessage, b'foo', b'foo']),
			1000*(e3.SynchronizeMessage.bytes() + 2*b'd\x00\x00\x00\x07foo'))
		pack_head = struct.Struct("!lH").pack
		# tuple
		self.assertEqual(e3.cat_messages([(b'foo',),]),
			b'D' + pack_head(7 + 4 + 2, 1) + b'\x00\x00\x00\x03foo')
		# tuple(foo,\N)
		self.assertEqual(e3.cat_messages([(b'foo',None,),]),
			b'D' + pack_head(7 + 4 + 4 + 2, 2) + b'\x00\x00\x00\x03foo\xFF\xFF\xFF\xFF')
		# tuple(foo,\N,bar)
		self.assertEqual(e3.cat_messages([(b'foo',None,b'bar'),]),
			b'D' + pack_head(7 + 7 + 4 + 4 + 2, 3) + \
			b'\x00\x00\x00\x03foo\xFF\xFF\xFF\xFF\x00\x00\x00\x03bar')
		# too many attributes
		self.assertRaises((OverflowError, struct.error),
			e3.cat_messages, [(None,) * 0x10000])

		class ThisEx(Exception):
			pass
		class ThatEx(Exception):
			pass
		class Bad(e3.Message):
			def serialize(self):
				raise ThisEx('foo')
		self.assertRaises(ThisEx, e3.cat_messages, [Bad()])
		class NoType(e3.Message):
			def serialize(self):
				return b''
		self.assertRaises(AttributeError, e3.cat_messages, [NoType()])
		class BadType(e3.Message):
			type = 123
			def serialize(self):
				return b''
		self.assertRaises((TypeError,struct.error), e3.cat_messages, [BadType()])


	def testSerializeParseConsistency(self):
		for msg in message_samples:
			smsg = msg.serialize()
			self.assertEqual(msg, msg.parse(smsg))

	def testEmptyMessages(self):
		for x in e3.__dict__.values():
			if isinstance(x, e3.EmptyMessage):
				xtype = type(x)
				self.assertTrue(x is xtype())

	def testUnknownNoticeFields(self):
		N = e3.Notice.parse(b'\x00\x00Z\x00Xklsvdnvldsvkndvlsn\x00Pfoobar\x00Mmessage\x00')
		E = e3.Error.parse(b'Z\x00Xklsvdnvldsvkndvlsn\x00Pfoobar\x00Mmessage\x00\x00')
		self.assertEqual(N[b'M'], b'message')
		self.assertEqual(E[b'M'], b'message')
		self.assertEqual(N[b'P'], b'foobar')
		self.assertEqual(E[b'P'], b'foobar')
		self.assertEqual(len(N), 4)
		self.assertEqual(len(E), 4)

	def testCompleteExtracts(self):
		x = e3.Complete(b'FOO BAR 1321')
		self.assertEqual(x.extract_command(), b'FOO BAR')
		self.assertEqual(x.extract_count(), 1321)
		x = e3.Complete(b' CREATE  	TABLE 13210  ')
		self.assertEqual(x.extract_command(), b'CREATE  	TABLE')
		self.assertEqual(x.extract_count(), 13210)
		x = e3.Complete(b'  CREATE  	TABLE  \t713210  ')
		self.assertEqual(x.extract_command(), b'CREATE  	TABLE')
		self.assertEqual(x.extract_count(), 713210)
		x = e3.Complete(b'  CREATE  	TABLE  0 \t13210  ')
		self.assertEqual(x.extract_command(), b'CREATE  	TABLE')
		self.assertEqual(x.extract_count(), 13210)
		x = e3.Complete(b' 0 \t13210 ')
		self.assertEqual(x.extract_command(), None)
		self.assertEqual(x.extract_count(), 13210)

##
# .protocol.xact3 tests
##

xact_samples = [
	# Simple contrived exchange.
	(
		(
			e3.Query(b"COMPLETE"),
		), (
			e3.Complete(b'COMPLETE'),
			e3.Ready(b'I'),
		)
	),
	(
		(
			e3.Query(b"ROW DATA"),
		), (
			e3.TupleDescriptor((
				(b'foo', 1, 1, 1, 1, 1, 1),
				(b'bar', 1, 2, 1, 1, 1, 1),
			)),
			e3.Tuple((b'lame', b'lame')),
			e3.Complete(b'COMPLETE'),
			e3.Ready(b'I'),
		)
	),
	(
		(
			e3.Query(b"ROW DATA"),
		), (
			e3.TupleDescriptor((
				(b'foo', 1, 1, 1, 1, 1, 1),
				(b'bar', 1, 2, 1, 1, 1, 1),
			)),
			e3.Tuple((b'lame', b'lame')),
			e3.Tuple((b'lame', b'lame')),
			e3.Tuple((b'lame', b'lame')),
			e3.Tuple((b'lame', b'lame')),
			e3.Ready(b'I'),
		)
	),
	(
		(
			e3.Query(b"NULL"),
		), (
			e3.Null(),
			e3.Ready(b'I'),
		)
	),
	(
		(
			e3.Query(b"COPY TO"),
		), (
			e3.CopyToBegin(1, [1,2]),
			e3.CopyData(b'row1'),
			e3.CopyData(b'row2'),
			e3.CopyDone(),
			e3.Complete(b'COPY TO'),
			e3.Ready(b'I'),
		)
	),
	(
		(
			e3.Function(1, [b''], [b''], 1),
		), (
			e3.FunctionResult(b'foo'),
			e3.Ready(b'I'),
		)
	),
	(
		(
			e3.Parse(b"NAME", b"SQL", ()),
		), (
			e3.ParseComplete(),
		)
	),
	(
		(
			e3.Bind(b"NAME", b"STATEMENT_ID", (), (), ()),
		), (
			e3.BindComplete(),
		)
	),
	(
		(
			e3.Parse(b"NAME", b"SQL", ()),
			e3.Bind(b"NAME", b"STATEMENT_ID", (), (), ()),
		), (
			e3.ParseComplete(),
			e3.BindComplete(),
		)
	),
	(
		(
			e3.Describe(b"STATEMENT_ID"),
		), (
			e3.AttributeTypes(()),
			e3.NoData(),
		)
	),
	(
		(
			e3.Describe(b"STATEMENT_ID"),
		), (
			e3.AttributeTypes(()),
			e3.TupleDescriptor(()),
		)
	),
	(
		(
			e3.CloseStatement(b"foo"),
		), (
			e3.CloseComplete(),
		),
	),
	(
		(
			e3.ClosePortal(b"foo"),
		), (
			e3.CloseComplete(),
		),
	),
	(
		(
			e3.Synchronize(),
		), (
			e3.Ready(b'I'),
		),
	),
]

class test_xact3(unittest.TestCase):
	def testTransactionSamplesAll(self):
		for xcmd, xres in xact_samples:
			x = x3.Instruction(xcmd)
			r = tuple([(y.type, y.serialize()) for y in xres])
			x.state[1]()
			self.assertEqual(x.messages, ())
			x.state[1](r)
			self.assertEqual(x.state, x3.Complete)
			rec = []
			for y in x.completed:
				for z in y[1]:
					if type(z) is type(b''):
						z = e3.CopyData(z)
					rec.append(z)
			self.assertEqual(xres, tuple(rec))

	def testClosing(self):
		c = x3.Closing()
		self.assertEqual(c.messages, (e3.DisconnectMessage,))
		c.state[1]()
		self.assertEqual(c.fatal, True)
		self.assertEqual(c.error_message.__class__, e3.ClientError)
		self.assertEqual(c.error_message[b'C'], '08003')

	def testNegotiation(self):
		# simple successful run
		n = x3.Negotiation({}, b'')
		n.state[1]()
		n.state[1](
			pairs(
				e3.Notice(((b'M', b"foobar"),)),
				e3.Authentication(e3.AuthRequest_OK, b''),
				e3.KillInformation(0,0),
				e3.ShowOption(b'name', b'val'),
				e3.Ready(b'I'),
			)
		)
		self.assertEqual(n.state, x3.Complete)
		self.assertEqual(n.last_ready.xact_state, b'I')
		# no killinfo.. should cause protocol error...
		n = x3.Negotiation({}, b'')
		n.state[1]()
		n.state[1](
			pairs(
				e3.Notice(((b'M', b"foobar"),)),
				e3.Authentication(e3.AuthRequest_OK, b''),
				e3.ShowOption(b'name', b'val'),
				e3.Ready(b'I'),
			)
		)
		self.assertEqual(n.state, x3.Complete)
		self.assertEqual(n.last_ready, None)
		self.assertEqual(n.error_message[b'C'], '08P01')
		# killinfo twice.. must cause protocol error...
		n = x3.Negotiation({}, b'')
		n.state[1]()
		n.state[1](
			pairs(
				e3.Notice(((b'M', b"foobar"),)),
				e3.Authentication(e3.AuthRequest_OK, b''),
				e3.ShowOption(b'name', b'val'),
				e3.KillInformation(0,0),
				e3.KillInformation(0,0),
				e3.Ready(b'I'),
			)
		)
		self.assertEqual(n.state, x3.Complete)
		self.assertEqual(n.last_ready, None)
		self.assertEqual(n.error_message[b'C'], '08P01')
		# start with ready message..
		n = x3.Negotiation({}, b'')
		n.state[1]()
		n.state[1](
			pairs(
				e3.Notice(((b'M', b"foobar"),)),
				e3.Ready(b'I'),
				e3.Authentication(e3.AuthRequest_OK, b''),
				e3.ShowOption(b'name', b'val'),
			)
		)
		self.assertEqual(n.state, x3.Complete)
		self.assertEqual(n.last_ready, None)
		self.assertEqual(n.error_message[b'C'], '08P01')
		# unsupported authreq
		n = x3.Negotiation({}, b'')
		n.state[1]()
		n.state[1](
			pairs(
				e3.Authentication(255, b''),
			)
		)
		self.assertEqual(n.state, x3.Complete)
		self.assertEqual(n.last_ready, None)
		self.assertEqual(n.error_message[b'C'], '--AUT')

	def testInstructionAsynchook(self):
		l = []
		def hook(data):
			l.append(data)
		x = x3.Instruction([
			e3.Query(b"NOTHING")
		], asynchook = hook)
		a1 = e3.Notice(((b'M', b"m1"),))
		a2 = e3.Notify(0, b'relation', b'parameter')
		a3 = e3.ShowOption(b'optname', b'optval')
		# "send" the query message
		x.state[1]()
		# "receive" the tuple
		x.state[1]([(a1.type, a1.serialize()),])
		a2l = [(a2.type, a2.serialize()),]
		x.state[1](a2l)
		# validate that the hook is not fed twice because
		# it's the exact same message set. (later assertion will validate)
		x.state[1](a2l)
		x.state[1]([(a3.type, a3.serialize()),])
		# we only care about validating that l got everything.
		self.assertEqual([a1,a2,a3], l)
		self.assertEqual(x.state[0], x3.Receiving)
		# validate that the asynchook exception is trapped.
		class Nee(Exception):
			pass
		def ehook(msg):
			raise Nee("this should **not** be part of the summary")
		x = x3.Instruction([
			e3.Query(b"NOTHING")
		], asynchook = ehook)
		a1 = e3.Notice(((b'M', b"m1"),))
		x.state[1]()
		import sys
		v = None
		def exchook(typ, val, tb):
			nonlocal v
			v = val
		seh = sys.excepthook
		sys.excepthook = exchook
		# we only care about validating that the exchook got called.
		x.state[1]([(a1.type, a1.serialize())])
		sys.excepthook = seh
		self.assertTrue(isinstance(v, Nee))

class test_client3(unittest.TestCase):
	def test_timeout(self):
		portnum = find_available_port()
		servsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		with servsock:
			servsock.bind(('localhost', portnum))
			pc = c3.Connection(
				SocketFactory(
					(socket.AF_INET, socket.SOCK_STREAM),
					('localhost', portnum)
				),
				{}
			)
			pc.connect(timeout = 1)
			try:
				self.assertEqual(pc.xact.fatal, True)
				self.assertEqual(pc.xact.__class__, x3.Negotiation)
			finally:
				if pc.socket is not None:
					pc.socket.close()

	def test_SSL_failure(self):
		portnum = find_available_port()
		servsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		with servsock:
			servsock.bind(('localhost', portnum))
			pc = c3.Connection(
				SocketFactory(
					(socket.AF_INET, socket.SOCK_STREAM),
					('localhost', portnum)
				),
				{}
			)
			exc = None
			servsock.listen(1)
			def client_thread():
				pc.connect(ssl = True)
			client = Thread(target = client_thread)
			try:
				client.start()
				c, addr = servsock.accept()
				with c:
					c.send(b'S')
					c.sendall(b'0000000000000000000000')
					c.recv(1024)
					c.close()
				client.join()
			finally:
				if pc.socket is not None:
					pc.socket.close()

		self.assertEqual(pc.xact.fatal, True)
		self.assertEqual(pc.xact.__class__, x3.Negotiation)
		self.assertEqual(pc.xact.error_message.__class__, e3.ClientError)
		self.assertTrue(hasattr(pc.xact, 'exception'))

	def test_bad_negotiation(self):
		portnum = find_available_port()
		servsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		servsock.bind(('localhost', portnum))
		pc = c3.Connection(
			SocketFactory(
				(socket.AF_INET, socket.SOCK_STREAM),
				('localhost', portnum)
			),
			{}
		)
		exc = None
		servsock.listen(1)
		def client_thread():
			pc.connect()
		client = Thread(target = client_thread)
		try:
			client.start()
			c, addr = servsock.accept()
			try:
				c.recv(1024)
			finally:
				c.close()
			time.sleep(0.25)
			client.join()
			servsock.close()
			self.assertEqual(pc.xact.fatal, True)
			self.assertEqual(pc.xact.__class__, x3.Negotiation)
			self.assertEqual(pc.xact.error_message.__class__, e3.ClientError)
			self.assertEqual(pc.xact.error_message[b'C'], '08006')
		finally:
			servsock.close()
			if pc.socket is not None:
				pc.socket.close()

if __name__ == '__main__':
	from types import ModuleType
	this = ModuleType("this")
	this.__dict__.update(globals())
	try:
		unittest.main(this)
	finally:
		import gc
		gc.collect()



class Varbit(object):
	__slots__ = ('data', 'bits')

	def from_bits(subtype, bits, data):
		if bits == 1:
			return (data[0] & (1 << 7)) and OneBit or ZeroBit
		else:
			rob = object.__new__(subtype)
			rob.bits = bits
			rob.data = data
			return rob
	from_bits = classmethod(from_bits)

	def __new__(typ, data):
		if isinstance(data, Varbit):
			return data
		if isinstance(data, bytes):
			return typ.from_bits(len(data) * 8, data)
		# str(), eg '00101100'
		bits = len(data)
		nbytes, remain = divmod(bits, 8)
		bdata = [bytes((int(data[x:x+8], 2),)) for x in range(0, bits - remain, 8)]
		if remain != 0:
			bdata.append(bytes((int(data[nbytes*8:].ljust(8,'0'), 2),)))
		return typ.from_bits(bits, b''.join(bdata))

	def __str__(self):
		if self.bits:
			# cut off the remainder from the bits
			blocks = [bin(x)[2:].rjust(8, '0') for x in self.data]
			blocks[-1] = blocks[-1][0:(self.bits % 8) or 8]
			return ''.join(blocks)
		else:
			return ''

	def __repr__(self):
		return '%s.%s(%r)' %(
			type(self).__module__,
			type(self).__name__,
			str(self)
		)

	def __eq__(self, ob):
		if not isinstance(ob, type(self)):
			ob = type(self)(ob)
		return ob.bits == self.bits and ob.data == self.data

	def __len__(self):
		return self.bits

	def __add__(self, ob):
		return Varbit(str(self) + str(ob))

	def __mul__(self, ob):
		return Varbit(str(self) * ob)

	def getbit(self, bitoffset):
		if bitoffset < 0:
			idx = self.bits + bitoffset
		else:
			idx = bitoffset
		if not 0 <= idx < self.bits:
			raise IndexError("bit index %d out of range" %(bitoffset,))

		byte, bitofbyte = divmod(idx, 8)
		if ord(self.data[byte]) & (1 << (7 - bitofbyte)):
			return OneBit
		else:
			return ZeroBit

	def __getitem__(self, item):
		if isinstance(item, slice):
			return type(self)(str(self)[item])
		else:
			return self.getbit(item)

	def __nonzero__(self):
		for x in self.data:
			if x != 0:
				return True
		return False

class Bit(Varbit):
	def __new__(subtype, ob):
		if ob is ZeroBit or ob is False or ob == '0':
			return ZeroBit
		elif ob is OneBit or ob is True or ob == '1':
			return OneBit

		raise ValueError('unknown bit value %r, 0 or 1' %(ob,))

	def __nonzero__(self):
		return self is OneBit

	def __str__(self):
		return self is OneBit and '1' or '0'

ZeroBit = object.__new__(Bit)
ZeroBit.data = b'\x00'
ZeroBit.bits = 1
OneBit = object.__new__(Bit)
OneBit.data = b'\x80'
OneBit.bits = 1



##
# .python.os
##
"""
General OS abstractions and information.
"""
import sys
import os

#: By default, close the FDs on subprocess.Popen().
close_fds = True

#: By default, there is no modification for executable references.
platform_exe = str

def find_file(basename, paths,
	join = os.path.join, exists = os.path.exists,
):
	"""
	Find the file in the given paths. Return the first path
	that exists.
	"""
	for x in paths:
		path = join(x, basename)
		if exists(path):
			return path

if sys.platform in ('win32','win64'):
	# replace variants for windows
	from .msw import close_fds, platform_exe

def find_executable(basename, pathsep = os.pathsep, platexe = platform_exe):
	paths = os.environ.get('PATH', '').split(pathsep)
	return find_file(platexe(basename), paths)



# -*- coding: utf-8 -*-

import hmac
from hashlib import md5, pbkdf2_hmac, sha256
from ..python.structlib import ulong_unpack

PLAIN_PASSWORD = 0
MD5_PASSWORD = 1
SHA256_PASSWORD = 2


def sha256_pw(user, password, salt):
	password_stored_method, salt = ulong_unpack(salt[:4]), salt[4:]
	if password_stored_method in (PLAIN_PASSWORD, SHA256_PASSWORD):
		random64_code_str, salt = salt[:64].decode(), salt[64:]
		token_str, salt = salt[:8].decode(), salt[8:]
		iteration = ulong_unpack(salt[:4])
		return rfc5802_algorithm(password, random64_code_str, token_str, "", iteration)
	elif password_stored_method == 1:
		# MD5
		pw = md5(password + user).hexdigest().encode('ascii')
		return b'md5' + md5(pw + salt[:4]).hexdigest().encode('ascii')
	else:
		raise Exception("pq: the password-stored method is not supported, must be plain, md5 or sha256.")


def rfc5802_algorithm(password, random64_code_str, token_str, server_signature, server_iteration):
	k = generate_k_from_pbkdf2(password, random64_code_str, server_iteration)
	server_key = get_key_from_hmac(k, b'Sever Key')
	client_key = get_key_from_hmac(k, b'Client Key')
	stored_key = get_sha256(client_key)
	token_bytes = hex_string_to_bytes(token_str)
	client_signature = get_key_from_hmac(server_key, token_bytes)
	if server_signature != "" and server_signature != bytes_to_hex_string(client_signature):
		return b''
	hmac_result = get_key_from_hmac(stored_key, token_bytes)
	h = XOR_between_password(hmac_result, client_key, len(client_key))
	return bytes_to_hex(h)


def XOR_between_password(password1, password2, length):
	arr = bytearray()
	for i in range(length):
		arr.append(password1[i] ^ password2[i])
	return bytes(arr)


def bytes_to_hex(bytes_array):
	lookup = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f']
	res = ""
	for i in range(len(bytes_array)):
		c = bytes_array[i] & 0xFF
		j = c >> 4
		res += (lookup[j])
		j = (c & 0xF)
		res += (lookup[j])
	return res.encode('ascii')


def get_key_from_hmac(k, key):
	h = hmac.new(k, digestmod=sha256)
	h.update(key)
	return h.digest()


def get_sha256(key):
	h = sha256()
	h.update(key)
	return h.digest()


def generate_k_from_pbkdf2(password, random64_code_str, iterations):
	random32_code = hex_string_to_bytes(random64_code_str)
	return pbkdf2_hmac('sha1', password, random32_code, iterations, dklen=32)


def hex_string_to_bytes(s):
	if not s:
		return b''

	arr = bytearray()
	s = s.upper()
	bytes_len = int(len(s) / 2)
	for i in range(bytes_len):
		pos = i * 2
		arr.append(ctoi(s[pos]) << 4 | ctoi(s[pos + 1]))
	return bytes(arr)


def bytes_to_hex_string(bs):
	s = ""
	for b in bs:
		v = b & 0xFF
		hv = "%x" % v
		if len(hv) < 2:
			s += hv
			s += "0"
		else:
			s += hv
	return s


def ctoi(c):
	return "0123456789ABCDEF".index(c)



##
# .types.io - I/O routines for packing and unpacking data
##
"""
PostgreSQL type I/O routines--packing and unpacking functions.

This package manages the modules providing I/O routines.

The name of the function describes what type the function is intended to be used
on. Normally, the fucntions return a structured form of the serialized data to
be used as a parameter to the creation of a higher level instance. In
particular, most of the functions that deal with time return a pair for
representing the relative offset: (seconds, microseconds). For times, this
provides an abstraction for quad-word based times used by some configurations of
PostgreSQL.
"""
import sys
from itertools import cycle, chain
from ... import types as pg_types

io_modules = {
	'builtins' : (
		pg_types.BOOLOID,
		pg_types.CHAROID,
		pg_types.BYTEAOID,

		pg_types.INT2OID,
		pg_types.INT4OID,
		pg_types.INT8OID,

		pg_types.FLOAT4OID,
		pg_types.FLOAT8OID,
		pg_types.ABSTIMEOID,
	),

	'pg_bitwise': (
		pg_types.BITOID,
		pg_types.VARBITOID,
	),

        'pg_network': (
		pg_types.MACADDROID,
		pg_types.INETOID,
		pg_types.CIDROID,
        ),

        'pg_system': (
		pg_types.OIDOID,
		pg_types.XIDOID,
		pg_types.CIDOID,
		pg_types.TIDOID,
	),

	'pg_geometry': (
		pg_types.POINTOID,
		pg_types.LSEGOID,
		pg_types.BOXOID,
		pg_types.CIRCLEOID,
	),

	'stdlib_datetime' : (
		pg_types.DATEOID,
		pg_types.INTERVALOID,
		pg_types.TIMEOID,
		pg_types.TIMETZOID,
		pg_types.TIMESTAMPOID,
		pg_types.TIMESTAMPTZOID
	),

	'stdlib_decimal' : (
		pg_types.NUMERICOID,
	),

	'stdlib_uuid' : (
		pg_types.UUIDOID,
	),

	'stdlib_xml_etree' : (
		pg_types.XMLOID,
	),

	'stdlib_jsonb' : (
		pg_types.JSONBOID,
	),

	# Must be db.typio.identify(contrib_hstore = 'hstore')'d
	'contrib_hstore' : (
		'contrib_hstore',
	),
}

# OID -> module name
module_io = dict(
	chain.from_iterable((
		zip(x[1], cycle((x[0],))) for x in io_modules.items()
	))
)

if sys.version_info[:2] < (3,3):
	def load(relmod):
		return __import__(__name__ + '.' + relmod, fromlist = True, level = 1)
else:
	def load(relmod):
		return __import__(relmod, globals = globals(), locals = locals(), fromlist = [''], level = 1)

def resolve(oid):
	io = module_io.get(oid)
	if io is None:
		return None
	if io.__class__ is str:
		module_io.update(load(io).oid_to_io)
		io = module_io[oid]
	return io



import uuid
from ...types import UUIDOID

def uuid_pack(x, UUID = uuid.UUID, bytes = bytes):
	if isinstance(x, UUID):
		return bytes(x.bytes)
	return bytes(UUID(x).bytes)

def uuid_unpack(x, UUID = uuid.UUID):
	return UUID(bytes=x)

oid_to_io = {
	UUIDOID : (uuid_pack, uuid_unpack),
}



"""
Split, unsplit, parse, serialize, construct and structure resource indicators.

Resource indicators take the form::

  [scheme:[//]][user[:pass]@]host[:port][/[path[/path]*][?param-n1=value[&param-n=value-n]*][#fragment]]

It might be an URL, URI, or IRI. It tries not to care.
Notably, it only percent-encodes chr(0-33) as some RIs support character values
greater than 127. Usually, it's best to make a second pass on the string in
order to target a specific format, URI or IRI.

If a specific format is being targeted, URL or URI or URI-represention of an
IRI, a second pass *must* be made on the string.
# Future versions may include subsequent transformation routines for targeting.

Overview
--------

Where ``x`` is a text RI(ie, ``http://foo.com/path``)::

	unsplit(split(x)) == x
	serialize(parse(x)) == x
	parse(x) == structure(split(x))
	construct(parse(x)) == split(x)


Substructure
------------

In some cases, an RI may have additional structure that needs to be extracted.
To do this, the ``fieldproc`` keyword is used on `split_netloc`, `structure`,
and `parse` functions.

The ``fieldproc`` keyword is a callable that takes a single argument and returns
the processed field. By default, ``fieldproc`` is the `unescape` function which
will decode percent escapes. This is not desirable when substructure exists
within an RI's component as it can create ambiguity about a token when a
percent encoded variant is decoded.
"""
import re
import copy

pct_encode = '%%%0.2X'.__mod__
unescaped = '%' + ''.join([chr(x) for x in range(0, 33)])

percent_escapes_re = re.compile('(%[0-9a-fA-F]{2,2})+')
escape_re = re.compile('[%s]' %(re.escape(unescaped),))
escape_user_re = re.compile('[%s]' %(re.escape(unescaped + ':@/?#'),))
escape_password_re = re.compile('[%s]' %(re.escape(unescaped + '@/?#'),))
escape_host_re = re.compile('[%s]' %(re.escape(unescaped + '/?#'),))
escape_port_re = re.compile('[%s]' %(re.escape(unescaped + '/?#'),))
escape_path_re = re.compile('[%s]' %(re.escape(unescaped + '/?#'),))
escape_query_key_re = re.compile('[%s]' %(re.escape(unescaped + '&=#'),))
escape_query_value_re = re.compile('[%s]' %(re.escape(unescaped + '&#'),))

percent_escapes = {}
for x in range(256):
	k = '%0.2X'.__mod__(x)
	percent_escapes[k] = x
	percent_escapes[k.lower()] = x
	percent_escapes[k[0].lower() + k[1]] = x
	percent_escapes[k[0] + k[1].lower()] = x

scheme_chars = '-.+0123456789'
del x

def unescape(x, mkval = chr):
	"""
	Substitute percent escapes with literal characters.
	"""
	nstr = type(x)('')
	if isinstance(x, str):
		mkval = chr
	pos = 0
	end = len(x)
	while pos != end:
		newpos = x.find('%', pos)
		if newpos == -1:
			nstr += x[pos:]
			break
		else:
			nstr += x[pos:newpos]

		val = percent_escapes.get(x[newpos+1:newpos+3])
		if val is not None:
			nstr += mkval(val)
			pos = newpos + 3
		else:
			nstr += '%'
			pos = newpos + 1
	return nstr

def re_pct_encode(m):
	return pct_encode(ord(m.group(0)))	

indexes = {
	'scheme' : 0,
	'netloc' : 1,
	'path' : 2,
	'query' : 3,
	'fragment' : 4
}

def split(s):
	"""
	Split an IRI into its base components based on the markers::

		://, /, ?, #

	Return a 5-tuple: (scheme, netloc, path, query, fragment)
	"""
	scheme = None
	netloc = None
	path = None
	query = None
	fragment = None

	end = len(s)
	pos = 0

	# Non-iauthority RI's should be special cased by the user.
	scheme_pos = s.find('://')
	if scheme_pos != -1:
		pos = scheme_pos + 3
		scheme = s[:scheme_pos]
		for x in scheme:
			if not (x in scheme_chars) and \
			not ('A' <= x <= 'Z') and not ('a' <= x <= 'z'):
				pos = 0
				scheme = None
				break

	end_of_netloc = end

	path_pos = s.find('/', pos)
	if path_pos == -1:
		path_pos = None
	else:
		end_of_netloc = path_pos

	query_pos = s.find('?', pos)
	if query_pos == -1:
		query_pos = None
	elif path_pos is None or query_pos < path_pos:
		path_pos = None
		end_of_netloc = query_pos

	fragment_pos = s.find('#', pos)
	if fragment_pos == -1:
		fragment_pos = None
	else:
		if query_pos is not None and fragment_pos < query_pos:
			query_pos = None
		if path_pos is not None and fragment_pos < path_pos:
			path_pos = None
			end_of_netloc = fragment_pos
		if query_pos is None and path_pos is None:
			end_of_netloc = fragment_pos

	if end_of_netloc != pos:
		netloc = s[pos:end_of_netloc]

	if path_pos is not None:
		path = s[path_pos+1:query_pos or fragment_pos or end]

	if query_pos is not None:
		query = s[query_pos+1:fragment_pos or end]

	if fragment_pos is not None:
		fragment = s[fragment_pos+1:end]

	return (scheme, netloc, path, query, fragment)

def unsplit_path(p, re = escape_path_re):
	"""
	Join a list of paths(strings) on "/" *after* escaping them.
	"""
	if not p:
		return None
	return '/'.join([re.sub(re_pct_encode, x) for x in p])

def split_path(p, fieldproc = unescape):
	"""
	Return a list of unescaped strings split on "/".

	Set `fieldproc` to `str` if the components' percent escapes should not be
	decoded.
	"""
	if p is None:
		return []
	return [fieldproc(x) for x in p.split('/')]

def unsplit(t):
	"""
	Make a RI from a split RI(5-tuple).
	"""
	s = ''
	if t[0] is not None:
		s += t[0]
		s += '://'
	if t[1] is not None:
		s += t[1]
	if t[2] is not None:
		s += '/'
		s += t[2]
	if t[3] is not None:
		s += '?'
		s += t[3]
	if t[4] is not None:
		s += '#'
		s += t[4]
	return s

def split_ip(ip, fieldproc = unescape):
	if ip[0] == '[':
		# IPvN addr
		next_pos = ip.find(']')
		if next_pos == -1:
			# unterminated IPvN block
			next_pos = len(ip) - 1
		addr = ip[:next_pos + 1]
		pos = next_pos + 1
		next_pos = addr.find(':', pos)
		if next_pos == -1:
			port = None
		else:
			port = fieldproc(addr[next_pos + 1:])
	else:
		next_pos = ip.find(':')
		if next_pos == -1:
			addr = fieldproc(ip)
			port = None
		else:
			addr = fieldproc(ip[:next_pos])
			port = fieldproc(ip[next_pos + 1:])
	return addr, port

def split_netloc(netloc, fieldproc = unescape):
	"""
	Split a net location into a 4-tuple list,
	[(user, password, host, port), (user, password, host, port), ...].

	Set `fieldproc` to `str` if the components' percent escapes should not be
	decoded.
	"""
	pos = netloc.find('@')
	if pos == -1:
		# No user information
		pos = 0
		user = None
		password = None
	else:
		s = netloc[:pos]
		userpw = s.split(':', 1)
		if len(userpw) == 2:
			user, password = userpw
			user = fieldproc(user)
			password = fieldproc(password)
		else:
			user = fieldproc(userpw[0])
			password = None
		pos += 1

	if pos >= len(netloc):
		return [(user, password, None, None)]

	res = []
	addr_start_pos = pos
	keep_going = True
	while keep_going:
		next_comma_pos = netloc.find(',', addr_start_pos)
		if next_comma_pos == -1:
			addr_end_pos = len(netloc)
			keep_going = False
		else:
			addr_end_pos = next_comma_pos
		addr, port = split_ip(netloc[addr_start_pos:addr_end_pos], fieldproc)
		res.append((user, password, addr, port))
		if keep_going:
			addr_start_pos = next_comma_pos+1
	return res

def unsplit_netloc(t):
	"""
	Create a netloc fragment from the given tuple(user,password,host,port).
	"""
	if t[0] is None and t[2] is None:
		return None
	s = ''
	if t[0] is not None:
		s += escape_user_re.sub(re_pct_encode, t[0])
		if t[1] is not None:
			s += ':'
			s += escape_password_re.sub(re_pct_encode, t[1])
		s += '@'

	if t[2] is not None:
		s += escape_host_re.sub(re_pct_encode, t[2])
		if t[3] is not None:
			s += ':'
			s += escape_port_re.sub(re_pct_encode, t[3])

	return s

def structure(t, fieldproc = unescape):
	"""
	Create a dictionary list from a split RI(5-tuple).

	Set `fieldproc` to `str` if the components' percent escapes should not be
	decoded.
	"""
	d = {}

	if t[0] is not None:
		d['scheme'] = t[0]

	if t[2] is not None:
		if t[2]:
			d['path'] = list(map(fieldproc, t[2].split('/')))
		else:
			d['path'] = []

	if t[3] is not None:
		if t[3]:
			d['query'] = [tuple((list(map(fieldproc, x.split('=', 1))) + [None])[:2]) for x in t[3].split('&')]
		else:
			# no characters followed the '?'
			d['query'] = []

	if t[4] is not None:
		d['fragment'] = fieldproc(t[4])

	if t[1] is None:  # netloc
		return [d]
	else:
		res = []
		uphps = split_netloc(t[1], fieldproc = fieldproc)
		for uphp in uphps:
			tmpd = copy.deepcopy(d)
			if uphp[0] is not None:
				tmpd['user'] = uphp[0]
			if uphp[1] is not None:
				tmpd['password'] = uphp[1]
			if uphp[2] is not None:
				tmpd['host'] = uphp[2]
			if uphp[3] is not None:
				tmpd['port'] = uphp[3]
			res.append(tmpd)
		return res

def construct_query(x,
	key_re = escape_query_key_re,
	value_re = escape_query_value_re,
):
	'Given a sequence of (key, value) pairs, construct'
	return '&'.join([
		v is not None and \
		'%s=%s' %(
			key_re.sub(re_pct_encode, k),
			value_re.sub(re_pct_encode, v),
		) or \
		key_re.sub(re_pct_encode, k)
		for k, v in x
	])

def construct(x):
	"""
	Construct a RI tuple(5-tuple) from a dictionary object.
	"""
	p = x.get('path')
	if p is not None:
		p = '/'.join([escape_path_re.sub(re_pct_encode, y) for y in p])
	q = x.get('query')
	if q is not None:
		q = construct_query(q)
	f = x.get('fragment')
	if f is not None:
		f = escape_re.sub(re_pct_encode, f)

	u = x.get('user')
	pw = x.get('password')
	h = x.get('host')
	port = x.get('port')

	return (
		x.get('scheme'),
		# netloc: [user[:pass]@]host[:port]
		unsplit_netloc((
			x.get('user'),
			x.get('password'),
			x.get('host'),
			x.get('port'),
		)),
		p, q, f
	)

def parse(s, fieldproc = unescape):
	"""
	Parse an RI into a dictionary object list. Synonym for ``structure(split(x))``.

	Set `fieldproc` to `str` if the components' percent escapes should not be
	decoded.
	"""
	return structure(split(s), fieldproc = fieldproc)

def serialize(x):
	"""
	Return an RI from a dictionary object. Synonym for ``unsplit(construct(x))``.
	"""
	return unsplit(construct(x))

__docformat__ = 'reStructuredText'



##
# .python.msw
##
"""
Additional Microsoft Windows tools.
"""

# for Popen(), not supported on windows
close_fds = False

def platform_exe(name):
	"""
	Append '.exe' if it's not already there.
	"""
	if name.endswith('.exe'):
		return name
	return name + '.exe'



from .. import \
	INT2OID, INT4OID, INT8OID, \
	BOOLOID, BYTEAOID, CHAROID, \
	ABSTIMEOID, FLOAT4OID, FLOAT8OID, \
	TEXTOID, BPCHAROID, NAMEOID, VARCHAROID
from . import lib

bool_pack = {True:b'\x01', False:b'\x00'}.__getitem__
bool_unpack = {b'\x01':True, b'\x00':False}.__getitem__

int2_pack, int2_unpack = lib.short_pack, lib.short_unpack
int4_pack, int4_unpack = lib.long_pack, lib.long_unpack
int8_pack, int8_unpack = lib.longlong_pack, lib.longlong_unpack

bytea_pack = bytes
bytea_unpack = bytes
char_pack = bytes
char_unpack = bytes

oid_to_io = {
	BOOLOID : (bool_pack, bool_unpack, bool),

	BYTEAOID : (bytea_pack, bytea_unpack, bytes),
	CHAROID : (char_pack, char_unpack, bytes),

	INT2OID : (int2_pack, int2_unpack, int),
	INT4OID : (int4_pack, int4_unpack, int),
	INT8OID : (int8_pack, int8_unpack, int),

	ABSTIMEOID : (lib.long_pack, lib.long_unpack, int),
	FLOAT4OID : (lib.float_pack, lib.float_unpack, float),
	FLOAT8OID : (lib.double_pack, lib.double_unpack, float),
}

# Python Representations of PostgreSQL Types
oid_to_type = {
	BOOLOID: bool,

	VARCHAROID: str,
	TEXTOID: str,
	BPCHAROID: str,
	NAMEOID: str,

	# This is *not* bpchar, the SQL CHARACTER type.
	CHAROID: bytes,
	BYTEAOID: bytes,

	INT2OID: int,
	INT4OID: int,
	INT8OID: int,

	FLOAT4OID: float,
	FLOAT8OID: float,
}



#!/usr/bin/env python
import sys
import os
from optparse import OptionParser
from .. import configfile
from .. import __version__

__all__ = ['command']

def command(args):
	"""
	pg_dotconf script entry point.
	"""
	op = OptionParser(
		"%prog [--stdout] [-f settings] postgresql.conf ([param=val]|[param])*",
		version = __version__
	)
	op.add_option(
		'-f', '--file',
		dest = 'settings',
		help = 'A file of settings to *apply* to the given "postgresql.conf"',
		default = [],
		action = 'append',
	)
	op.add_option(
		'--stdout',
		dest = 'stdout',
		help = 'Redirect the product to standard output instead of writing back to the "postgresql.conf" file',
		action = 'store_true',
		default = False
	)
	co, ca = op.parse_args(args[1:])
	if not ca:
		return 0

	settings = {}
	for sfp in co.settings:
		with open(sfp) as sf:
			for line in sf:
				pl = configfile.parse_line(line)
				if pl is not None:
					if comment not in line[pl[0].start]:
						settings[line[pl[0]]] = unquote(line[pl[1]])

	prev = None
	for p in ca[1:]:
		if '=' not in p:
			k = p
			v = None
		else:
			k, v = p.split('=', 1)
		k = k.strip()
		if not k:
			sys.stderr.write("ERROR: invalid setting, %r after %r%s" %(
				p, prev, os.linesep
			))
			sys.stderr.write(
				"HINT: Settings must take the form 'setting=value' " \
				"or 'setting_name_to_comment'. Settings must also be received " \
				"as a single argument." + os.linesep
			)
			sys.exit(1)
		prev = p
		settings[k] = v

	fp = ca[0]
	with open(fp, 'r') as fr:
		lines = configfile.alter_config(settings, fr)

	if co.stdout or fp == '/dev/stdin':
		for l in lines:
			sys.stdout.write(l)
	else:
		with open(fp, 'w') as fw:
			for l in lines:
				fw.write(l)
	return 0

if __name__ == '__main__':
	sys.exit(command(sys.argv))



##
# .test.cursor_integrity
##
import os
import unittest
import random
import itertools

iot = '_dst'

getq = "SELECT i FROM generate_series(0, %d) AS g(i)"
copy = "COPY (%s) TO STDOUT"

def random_read(curs, remaining_rows):
	"""
	Read from one of the three methods using a random amount if sized.
	- 50% chance of curs.read(random())
	- 40% chance of next()
	- 10% chance of read() # no count
	"""
	if random.random() > 0.5:
		rrows = random.randrange(0, remaining_rows)
		return curs.read(rrows), rrows
	elif random.random() < 0.1:
		return curs.read(), -1
	else:
		try:
			return [next(curs)], 1
		except StopIteration:
			return [], 1

def random_select_get(limit):
	return prepare(getq %(limit - 1,))

def random_copy_get(limit):
	return prepare(copy %(getq %(limit - 1,),))

class test_integrity(unittest.TestCase):
	"""
	test the integrity of the get and put interfaces on queries
	and result handles.
	"""
	def test_select(self):
		total = 0
		while total < 10000:
			limit = random.randrange(500000)
			read = 0
			total += limit
			p = random_select_get(limit)()
			last = ([(-1,)], 1)
			completed = [last[0]]
			while True:
				next = random_read(p, (limit - read) or 10)
				thisread = len(next[0])
				read += thisread
				completed.append(next[0])
				if thisread:
					self.failUnlessEqual(
						last[0][-1][0], next[0][0][0] - 1,
						"first row(-1) of next failed to match the last row of the previous"
					)
					last = next
				elif next[1] != 0:
					# done
					break
			self.failUnlessEqual(read, limit)
			self.failUnlessEqual(list(range(-1, limit)), [
				x[0] for x in itertools.chain(*completed)
			])

	def test_insert(self):
		pass

	if 'db' in dir(__builtins__) and pg.version_info >= (8,2,0):
		def test_copy_out(self):
			total = 0
			while total < 10000000:
				limit = random.randrange(500000)
				read = 0
				total += limit
				p = random_copy_get(limit)()
				last = ([-1], 1)
				completed = [last[0]]
				while True:
					next = random_read(p, (limit - read) or 10)
					next = ([int(x) for x in next[0]], next[1])
					thisread = len(next[0])
					read += thisread
					completed.append(next[0])
					if thisread:
						self.failUnlessEqual(
							last[0][-1], next[0][0] - 1,
							"first row(-1) of next failed to match the last row of the previous"
						)
						last = next
					elif next[1] != 0:
						# done
						break
				self.failUnlessEqual(read, limit)
				self.failUnlessEqual(
					list(range(-1, limit)),
					list(itertools.chain(*completed))
				)

	def test_copy_in(self):
		pass

def main():
	global copyin, loadin
	execute("CREATE TEMP TABLE _dst (i bigint)")
	copyin = prepare("COPY _dst FROM STDIN")
	loadin = prepare("INSERT INTO _dst VALUES ($1)")
	unittest.main()

if __name__ == '__main__':
	main()



##
# .protocol.version
##
"""
PQ version class used by startup messages.
"""
from struct import Struct
version_struct = Struct('!HH')

class Version(tuple):
	"""
	Version((major, minor)) -> Version

	Version serializer and parser.
	"""
	major = property(fget = lambda s: s[0])
	minor = property(fget = lambda s: s[1])

	def __new__(subtype, major_minor):
		(major, minor) = major_minor
		major = int(major)
		minor = int(minor)
		# If it can't be packed like this, it's not a valid version.
		try:
			version_struct.pack(major, minor)
		except Exception as e:
			raise ValueError("unpackable major and minor") from e

		return tuple.__new__(subtype, (major, minor))

	def __int__(self):
		return (self[0] << 16) | self[1]

	def bytes(self):
		return version_struct.pack(self[0], self[1])

	def __repr__(self):
		return '%d.%d' %(self[0], self[1])

	def parse(self, data):
		return self(version_struct.unpack(data))
	parse = classmethod(parse)

CancelRequestCode = Version((1234, 5678))
NegotiateSSLCode = Version((1234, 5679))
V2_0 = Version((2, 0))
V3_0 = Version((3, 0))
V3_51 = Version((3, 51)) # add for openGauss



##
# stdlib_datetime - support for the stdlib's datetime.
#
# I/O routines for date, time, timetz, timestamp, timestamptz, and interval.
# Supported by the datetime module.
##
import datetime
import warnings
from functools import partial
from operator import methodcaller, add

from ...python.datetime import UTC, FixedOffset, \
	infinity_date, infinity_datetime, \
	negative_infinity_date, negative_infinity_datetime
from ...python.functools import Composition as compose
from ...exceptions import TypeConversionWarning

from .. import \
	DATEOID, INTERVALOID, \
	TIMEOID, TIMETZOID, \
	TIMESTAMPOID, TIMESTAMPTZOID

from . import lib

oid_to_type = {
	DATEOID: datetime.date,
	TIMESTAMPOID: datetime.datetime,
	TIMESTAMPTZOID: datetime.datetime,
	TIMEOID: datetime.time,
	TIMETZOID: datetime.time,

	# XXX: datetime.timedelta doesn't support months.
	INTERVALOID: datetime.timedelta,
}

seconds_in_day = 24 * 60 * 60
seconds_in_hour = 60 * 60

pg_epoch_datetime = datetime.datetime(2000, 1, 1)
pg_epoch_datetime_utc = pg_epoch_datetime.replace(tzinfo = UTC)
pg_epoch_date = pg_epoch_datetime.date()
pg_date_offset = pg_epoch_date.toordinal()

## Difference between PostgreSQL epoch and Unix epoch.
## Used to convert a PostgreSQL ordinal to an ordinal usable by datetime
pg_time_days = (pg_date_offset - datetime.date(1970, 1, 1).toordinal())

##
# Constants used to special case infinity and -infinity.
time64_pack_constants = {
	infinity_datetime: lib.time64_infinity,
	negative_infinity_datetime: lib.time64_negative_infinity,
	'infinity': lib.time64_infinity,
	'-infinity': lib.time64_negative_infinity,
}
time_pack_constants = {
	infinity_datetime: lib.time_infinity,
	negative_infinity_datetime: lib.time_negative_infinity,
	'infinity': lib.time_infinity,
	'-infinity': lib.time_negative_infinity,
}
date_pack_constants = {
	infinity_date: lib.date_infinity,
	negative_infinity_date: lib.date_negative_infinity,
	'infinity': lib.date_infinity,
	'-infinity': lib.date_negative_infinity,
}
time64_unpack_constants = {
	lib.time64_infinity: infinity_datetime,
	lib.time64_negative_infinity: negative_infinity_datetime,
}
time_unpack_constants = {
	lib.time_infinity: infinity_datetime,
	lib.time_negative_infinity: negative_infinity_datetime,
}
date_unpack_constants = {
	lib.date_infinity: infinity_date,
	lib.date_negative_infinity: negative_infinity_date,
}

def date_pack(x,
	pack = lib.date_pack,
	offset = pg_date_offset,
	get = date_pack_constants.get,
):
	return get(x) or pack(x.toordinal() - offset)

def date_unpack(x,
	unpack = lib.date_unpack,
	offset = pg_date_offset,
	from_ord = datetime.date.fromordinal,
	get = date_unpack_constants.get,
):
	return get(x) or from_ord(unpack(x) + pg_date_offset)

def timestamp_pack(x,
	seconds_in_day = seconds_in_day,
	pg_epoch_datetime = pg_epoch_datetime,
):
	"""
	Create a (seconds, microseconds) pair from a `datetime.datetime` instance.
	"""
	x = (x - pg_epoch_datetime)
	return ((x.days * seconds_in_day) + x.seconds, x.microseconds)

def timestamp_unpack(seconds,
	timedelta = datetime.timedelta,
	relative_to = pg_epoch_datetime.__add__,
):
	"""
	Create a `datetime.datetime` instance from a (seconds, microseconds) pair.
	"""
	return relative_to(timedelta(0, *seconds))

def timestamptz_pack(x,
	seconds_in_day = seconds_in_day,
	pg_epoch_datetime_utc = pg_epoch_datetime_utc,
	UTC = UTC,
):
	"""
	Create a (seconds, microseconds) pair from a `datetime.datetime` instance.
	"""
	x = (x.astimezone(UTC) - pg_epoch_datetime_utc)
	return ((x.days * seconds_in_day) + x.seconds, x.microseconds)

def timestamptz_unpack(seconds,
	timedelta = datetime.timedelta,
	relative_to = pg_epoch_datetime_utc.__add__,
):
	"""
	Create a `datetime.datetime` instance from a (seconds, microseconds) pair.
	"""
	return relative_to(timedelta(0, *seconds))

def time_pack(x, seconds_in_hour = seconds_in_hour):
	"""
	Create a (seconds, microseconds) pair from a `datetime.time` instance.
	"""
	return (
		(x.hour * seconds_in_hour) + (x.minute * 60) + x.second,
		x.microsecond
	)

def time_unpack(seconds_ms, time = datetime.time, divmod = divmod):
	"""
	Create a `datetime.time` instance from a (seconds, microseconds) pair.
	Seconds being offset from epoch.
	"""
	seconds, ms = seconds_ms
	minutes, sec = divmod(seconds, 60)
	hours, min = divmod(minutes, 60)
	return time(hours, min, sec, ms)

def interval_pack(x):
	"""
	Create a (months, days, (seconds, microseconds)) tuple from a
	`datetime.timedelta` instance.
	"""
	return (0, x.days, (x.seconds, x.microseconds))

def interval_unpack(mds, timedelta = datetime.timedelta):
	"""
	Given a (months, days, (seconds, microseconds)) tuple, create a
	`datetime.timedelta` instance.
	"""
	months, days, seconds_ms = mds
	if months != 0:
		# XXX: Should this raise an exception?
		w = TypeConversionWarning(
			"datetime.timedelta cannot represent relative intervals",
			details = {
				'hint': 'An interval was unpacked with a non-zero "month" field.'
			},
			source = 'DRIVER'
		)
		warnings.warn(w)
	return timedelta(
		days = days + (months * 30),
		seconds = seconds_ms[0], microseconds = seconds_ms[1]
	)

def timetz_pack(x,
	time_pack = time_pack,
):
	"""
	Create a ((seconds, microseconds), timezone) tuple from a `datetime.time`
	instance.
	"""
	td = x.tzinfo.utcoffset(x)
	seconds = (td.days * seconds_in_day + td.seconds)
	return (time_pack(x), seconds)

def timetz_unpack(tstz,
	time_unpack = time_unpack,
	FixedOffset = FixedOffset,
):
	"""
	Create a `datetime.time` instance from a ((seconds, microseconds), timezone)
	tuple.
	"""
	t = time_unpack(tstz[0])
	return t.replace(tzinfo = FixedOffset(tstz[1]))

FloatTimes = False
IntTimes = True
NoDay = True
WithDay = False

# Used to handle the special cases: infinity and -infinity.
def proc_when_not_in(proc, dict):
	def _proc(x, get=dict.get):
		return get(x) or proc(x)
	return _proc

id_to_io = {
	(FloatTimes, TIMEOID) : (
		compose((time_pack, lib.time_pack)),
		compose((lib.time_unpack, time_unpack)),
		datetime.time
	),
	(FloatTimes, TIMETZOID) : (
		compose((timetz_pack, lib.timetz_pack)),
		compose((lib.timetz_unpack, timetz_unpack)),
		datetime.time
	),
	(FloatTimes, TIMESTAMPOID) : (
		proc_when_not_in(compose((timestamp_pack, lib.time_pack)), time_pack_constants),
		proc_when_not_in(compose((lib.time_unpack, timestamp_unpack)), time_unpack_constants),
		datetime.datetime
	),
	(FloatTimes, TIMESTAMPTZOID) : (
		proc_when_not_in(compose((timestamptz_pack, lib.time_pack)), time_pack_constants),
		proc_when_not_in(compose((lib.time_unpack, timestamptz_unpack)), time_unpack_constants),
		datetime.datetime
	),
	(FloatTimes, WithDay, INTERVALOID): (
		compose((interval_pack, lib.interval_pack)),
		compose((lib.interval_unpack, interval_unpack)),
		datetime.timedelta
	),
	(FloatTimes, NoDay, INTERVALOID): (
		compose((interval_pack, lib.interval_noday_pack)),
		compose((lib.interval_noday_unpack, interval_unpack)),
		datetime.timedelta
	),

	(IntTimes, TIMEOID) : (
		compose((time_pack, lib.time64_pack)),
		compose((lib.time64_unpack, time_unpack)),
		datetime.time
	),
	(IntTimes, TIMETZOID) : (
		compose((timetz_pack, lib.timetz64_pack)),
		compose((lib.timetz64_unpack, timetz_unpack)),
		datetime.time
	),
	(IntTimes, TIMESTAMPOID) : (
		proc_when_not_in(compose((timestamp_pack, lib.time64_pack)), time64_pack_constants),
		proc_when_not_in(compose((lib.time64_unpack, timestamp_unpack)), time64_unpack_constants),
		datetime.datetime
	),
	(IntTimes, TIMESTAMPTZOID) : (
		proc_when_not_in(compose((timestamptz_pack, lib.time64_pack)), time64_pack_constants),
		proc_when_not_in(compose((lib.time64_unpack, timestamptz_unpack)), time64_unpack_constants),
		datetime.datetime
	),
	(IntTimes, WithDay, INTERVALOID) : (
		compose((interval_pack, lib.interval64_pack)),
		compose((lib.interval64_unpack, interval_unpack)),
		datetime.timedelta
	),
	(IntTimes, NoDay, INTERVALOID) : (
		compose((interval_pack, lib.interval64_noday_pack)),
		compose((lib.interval64_noday_unpack, interval_unpack)),
		datetime.timedelta
	),
}

##
# Identify whether it's IntTimes or FloatTimes
def time_type(typio):
	idt = typio.database.settings.get('integer_datetimes', None)
	if idt is None:
		# assume its absence means its on after 9.0
		return bool(typio.database.version_info >= (9,0))
	elif idt.__class__ is bool:
		return idt
	else:
		return (idt.lower() in ('on', 'true', 't', True))

def select_format(oid, typio, get = id_to_io.__getitem__):
	return get((time_type(typio), oid))

def select_day_format(oid, typio, get = id_to_io.__getitem__):
	return get((time_type(typio), typio.database.version_info[:2] <= (8,0), oid))

oid_to_io = {
	DATEOID : (date_pack, date_unpack, datetime.date,),
	TIMEOID : select_format,
	TIMETZOID : select_format,
	TIMESTAMPOID : select_format,
	TIMESTAMPTZOID : select_format,
	INTERVALOID : select_day_format,
}



##
# .test.test_pgpassfile
##
import unittest
from .. import pgpassfile as client_pgpass
from io import StringIO

passfile_sample = """
# host:1111:dbname:user:password1
host:1111:dbname:user:password1
*:1111:dbname:user:password2
*:*:dbname:user:password3

# Comment

*:*:*:user:password4
*:*:*:usern:password4.5
*:*:*:*:password5
"""

passfile_sample_map = {
	('user', 'host', '1111', 'dbname') : 'password1',
	('user', 'host', '1111', 'dbname') : 'password1',
	('user', 'foo', '1111', 'dbname') : 'password2',
	('user', 'foo', '4321', 'dbname') : 'password3',
	('user', 'foo', '4321', 'db,name') : 'password4',

	('uuser', 'foo', '4321', 'db,name') : 'password5',
	('usern', 'foo', '4321', 'db,name') : 'password4.5',
	('foo', 'bar', '19231', 'somedbn') : 'password5',
}

difficult_passfile_sample = r"""
host\\:1111:db\:name:u\\ser:word1
*:1111:\:dbname\::\\user\\:pass\:word2
foohost:1111:\:dbname\::\\user\\:pass\:word3
"""

difficult_passfile_sample_map = {
	('u\\ser','host\\','1111','db:name') : 'word1',
	('\\user\\','somehost','1111',':dbname:') : 'pass:word2',
	('\\user\\','someotherhost','1111',':dbname:') : 'pass:word2',
	# More specific, but comes after '*'
	('\\user\\','foohost','1111',':dbname:') : 'pass:word2',
	('','','','') : None,
}

class test_pgpass(unittest.TestCase):
	def runTest(self):
		sample1 = client_pgpass.parse(StringIO(passfile_sample))
		sample2 = client_pgpass.parse(StringIO(difficult_passfile_sample))

		for k, pw in passfile_sample_map.items():
			lpw = client_pgpass.lookup_password(sample1, k)
			self.assertEqual(lpw, pw,
				"password lookup incongruity, expecting %r got %r with %r"
				" in \n%s" %(
					pw, lpw, k, passfile_sample
				)
			)

		for k, pw in difficult_passfile_sample_map.items():
			lpw = client_pgpass.lookup_password(sample2, k)
			self.assertEqual(lpw, pw,
				"password lookup incongruity, expecting %r got %r with %r"
				" in \n%s" %(
					pw, lpw, k, difficult_passfile_sample
				)
			)

if __name__ == '__main__':
	unittest.main()



##
# .lib - libraries; manage SQL outside of Python.
##
"""
PostgreSQL statement and object libraries.

The purpose of a library is provide a means to manage a mapping of symbols
to database operations or objects. These operations can be simple statements,
procedures, or something more complex.

Libraries are intended to allow the programmer to isolate and manage SQL outside
of a system's code-flow. It provides a means to construct the basic Python
interfaces to a PostgreSQL-based application.
"""
import io
import os.path
from types import ModuleType
from abc import abstractmethod, abstractproperty
from ..python.element import Element, ElementSet
from .. import api as pg_api
from .. import sys as pg_sys
from .. import exceptions as pg_exc
from ..python.itertools import find
from itertools import chain

try:
	libdir = os.path.abspath(os.path.dirname(__file__))
except NameError:
	pass
else:
	if os.path.exists(libdir):
		pg_sys.libpath.insert(0, libdir)
	del libdir

__all__ = [
	'Library',
	'SymbolCollection',
	'ILF',
	'Symbol',
	'Binding',
	'BoundSymbol',
	'find_libsql',
	'load',
]

class Symbol(Element):
	"""
	An annotated SQL statement string.

	The annotations describe how the statement should be used.
	"""
	__slots__ = (
		'library',
		'source',
		'name',
		'method',
		'type',
		'parameters',
	)
	_e_label = 'SYMBOL'
	_e_factors = ('library', 'source',)

	# The statement execution methods; symbols allow this to be specified
	# in order for a default method to be selected.
	execution_methods = {
		'first',
		'rows',
		'chunks',
		'declare',
		'load_chunks',
		'load_rows',
		'column',
	}

	def _e_metas(self):
		yield (None, self.name)

	def __init__(self,
		library, source,
		name = None,
		method = None,
		type = None,
		parameters = None,
		reference = False,
	):
		self.library = library
		self.source = source
		self.name = name
		if method in (None, '', 'all'):
			method = None
		elif method not in self.execution_methods:
			raise ValueError("unknown execution method: " + repr(method))
		self.method = method
		self.type = type
		self.parameters = parameters
		self.reference = reference

	def __str__(self):
		"""
		Provide the source of the query's symbol.
		"""
		# Explicitly run str() on source as it is expected that a
		# given symbol's source may be generated.
		return str(self.source)

class Library(Element):
	"""
	A library is mapping of symbol names to `postgresql.lib.Symbol` instances.
	"""
	_e_label = 'LIBRARY'
	_e_factors = ()

	@abstractproperty
	def address(self) -> str:
		"""
		A string indicating the source of the symbols.
		"""

	@abstractproperty
	def name(self) -> str:
		"""
		The name to bind the library as. Should be an identifier.
		"""

	@abstractproperty
	def preload(self) -> {str,}:
		"""
		A set of symbols that should prepared when the library is bound.
		"""

	@abstractmethod
	def symbols(self) -> [str]:
		"""
		Iterable of symbol names provides by the library.
		"""

	@abstractmethod
	def get_symbol(self, name) -> (Symbol, [Symbol]):
		"""
		Return the symbol with the given name.
		"""

class SymbolCollection(Library):
	"""
	Explicitly composed library. (Symbols passed into __init__)
	"""
	preload = None
	symtypes = (
		'static',
		'preload',
		'const',
		'proc',
		'transient',
	)

	def __init__(self, symbols, preface = None):
		"""
		Given an iterable of (symtype, symexe, doc, sql) tuples, create a
		symbol collection.
		"""
		self.preface = preface
		self._address = None
		self._name = None
		s = self.symbolsd = {}
		self.preload = set()
		for name, (isref, typ, exe, doc, query) in symbols:
			if typ and typ not in self.symtypes:
				raise ValueError(
					"symbol %r has an invalid type: %r" %(name, typ)
				)
			if typ == 'preload':
				self.preload.add(name)
				typ = None
			elif typ == 'proc':
				pass
			SYM = Symbol(self, query,
				name = name,
				method = exe,
				type = typ,
				reference = isref
			)
			s[name] = SYM

class ILF(SymbolCollection):
	'INI Library Format'
	def _e_metas(self):
		yield (None, self._address or 'ILF')

	def __repr__(self):
		return self.__class__.__module__ + '.' + self.__class__.__name__ + '.open(' + repr(self.address) + ')'

	@property
	def name(self):
		return self._name

	@property
	def address(self):
		return self._address

	def get_symbol(self, name):
		return self.symbolsd.get(name)

	def symbols(self):
		return self.symbolsd.keys()

	@classmethod
	def from_lines(typ, lines):
		"""
		Create an anonymous ILF library from a sequence of lines.
		"""
		prev = ''
		curid = None
		curblock = []
		blocks = []
		for line in lines:
			l = line.strip()
			if l.startswith('[') and l.endswith(']'):
				blocks.append((curid, curblock))
				curid = line
				curblock = []
			elif line.startswith('*[') and ']' in line:
				ref, rest = line.split(']', 1)
				# strip the leading '*['
				ref = ref[2:]
				# dereferencing will take place later.
				curblock.append((ref, rest))
			else:
				curblock.append(line)
		blocks.append((curid, curblock))
		preface = ''.join(blocks.pop(0)[1])
		syms = []
		for symdesc, block in blocks:
			# symbol name
			# symbol type
			# how to execute symbol
			name, styp, exe, *_ = (tuple(
				symdesc.strip().strip('[]').split(':')
			) + (None, None))
			doc = ''
			endofcomment = 0
			# resolve any symbol references; only one per line.
			block = [
				x if x.__class__ is not tuple else (
					find(reversed(syms), lambda y: y[0] == x[0])[1][-1] + x[1]
				)
				for x in block
			]
			for x in block:
				if x.startswith('-- '):
					doc += x[3:]
				else:
					break
				endofcomment += 1
			query = ''.join(block[endofcomment:])
			if styp == 'proc':
				query = query.strip()
			if name.startswith('&'):
				name = name[1:]
				isref = True
			else:
				isref = False
			syms.append((name, (isref, styp, exe, doc, query)))
		return typ(syms, preface = preface)

	@classmethod
	def open(typ, filepath, *args, **kw):
		"""
		Create a named ILF library from a file path.
		"""
		with io.open(filepath, *args, **kw) as fp:
			r = typ.from_lines(fp)
			r._address = os.path.abspath(filepath)
			bn = os.path.basename(filepath)
			if bn.startswith('lib') and bn.endswith('.sql'):
				r._name = bn[3:-4] or None
		return r

class BoundSymbol(object):
	"""
	A symbol bound to a database(connection).
	"""
	def __init__(self, symbol, database):
		if symbol.type == 'proc':
			proc = database.proc(symbol)
			self.method = proc.__call__
			self.object = proc
		else:
			ps = database.prepare(symbol)
			m = symbol.method
			if m is None:
				self.method = ps.__call__
			else:
				self.method = getattr(ps, m)
			self.object = ps

	def __call__(self, *args, **kw):
		return self.method(*args, **kw)

class BoundReference(object):
	"""
	A symbol bound to a database whose results make up the source of a symbol
	that will be created upon the execution of this symbol.

	A reference to a symbol.
	"""

	def __init__(self, symbol, database):
		self.symbol = symbol
		self.database = database
		self.method = database.prepare(symbol).chunks

	def __call__(self, *args, **kw):
		chunks = chain.from_iterable(self.method(*args, **kw))
		# Join columns with a space, and rows with a newline.
		src = '\n'.join([' '.join(row) for row in chunks])
		return BoundSymbol(
			Symbol(
				self.symbol.library, src,
				name = self.symbol.name,
				method = self.symbol.method,
				type = self.symbol.type,
				parameters = self.symbol.parameters,
				reference = False,
			),
			self.database,
		)

class Binding(object):
	"""
	Library bound to a database(connection).
	"""
	def __init__(self, database, library):
		self.__dict__.update({
			'__database__' : database,
			'__symbol_library__' : library,
			'__symbol_cache__' : {},
		})
		for x in library.preload:
			# cache all preloaded symbols.
			getattr(self, x)

	def __repr__(self):
		return '<Binding: lib%s on %r>' %(
			self.__symbol_library__.name,
			self.__database__
		)

	def __dir__(self):
		return dir(super()) + list(self.__symbol_library__.symbols())

	def __getattr__(self, name):
		"""
		Return a BoundSymbol against the Binding's database with the
		symbol named ``name`` in the Binding's library.
		"""
		d = self.__dict__
		s = d['__symbol_cache__']
		db = d['__database__']
		lib = d['__symbol_library__']

		bs = s.get(name)
		if bs is None:
			# No symbol cached with that name.
			# Everything is crammed in here because
			# we do *not* want methods on this object.
			# The namespace is primarily reserved for symbols.
			sym = lib.get_symbol(name)
			if sym is None:
				raise AttributeError(
					"symbol %r does not exist in library %r" %(
						name, lib.address
					)
				)
			if sym.reference:
				# Reference.
				bs = BoundReference(sym, db)
				if sym.type == 'const':
					# Constant Reference means a BoundSymbol.
					bs = bs()
				if sym.type != 'transient':
					s[name] = bs
			else:
				if not isinstance(sym, Symbol):
					# subjective symbol...
					sym = sym(db)
					if not isinstance(sym, Symbol):
						raise TypeError(
							"callable symbol, %r, did not produce " \
							"Symbol instance" %(name,)
						)
				if sym.type == 'const':
					r = BoundSymbol(sym, db)()
					if sym.method in ('chunks', 'rows', 'column'):
						# resolve the iterator
						r = list(r)
					bs = s[name] = r
				else:
					bs = BoundSymbol(sym, db)
					if sym.type != 'transient':
						s[name] = bs
		return bs

class Category(pg_api.Category):
	"""
	Library-based Category.
	"""
	_e_factors = ('libraries',)
	def _e_metas(self):
		yield ('aliases', {k.name: v for k, v in self.aliases.items()})

	def __init__(self, *libs, **named_libs):
		sl = set(libs)
		nl = set(named_libs.values())
		self._direct = sl
		self.libraries = ElementSet(sl | nl)
		self.aliases = {}
		# lib -> [alias-1, alias-2, ..., alias-n]
		for k, v in named_libs.items():
			d = self.aliases.setdefault(v, [])
			d.append(k)

	def __call__(self, database):
		for l in self.libraries:
			names = list(self.aliases.get(l, ()))
			if l in self._direct:
				names.append(l.name)
			B = Binding(database, l)
			for n in names:
				if hasattr(database, n):
					raise AttributeError("attribute already exists: " + name)
				setattr(database, n, B)

def find_libsql(libname, paths, prefix = 'lib', suffix = '.sql'):
	"""
	Given the base library name, `libname`, look for a file named
	"<prefix><libname><suffix>" in each directory(`paths`).
	All finds will be yielded out.
	"""
	lib = prefix + libname + suffix
	for p in paths:
		p = os.path.join(p, lib)
		if os.path.exists(p):
			yield p

def load(libref):
	"""
	Given a reference to a symbol library, instantiate the Library instance.

	Currently this function accepts:

	 * `str` objects as absolute paths or relative to sys.libpath.
	 * Module objects.
	"""
	if isinstance(libref, ModuleType):
		if hasattr(libref, '__lib'):
			lib = getattr(libref, '__lib')
		else:
			lib = ModuleLibrary(libref)
			setattr(libref, '__lib', lib)
	elif isinstance(libref, str):
		try:
			if os.path.sep in libref:
				# sep in libref? it's treated as a path.
				lib = ILF.open(libref)
			else:
				# first one wins.
				for x in find_libsql(libref, pg_sys.libpath):
					break
				else:
					raise pg_exc.LoadError("library %r not in postgresql.sys.libpath" % (libref,))
				lib = ILF.open(x)
		except pg_exc.LoadError:
			raise
		except Exception:
			# any exception is a load error.
			raise pg_exc.LoadError("failed load ILF, " + repr(libref))
	else:
		raise TypeError("load takes a module or str, given " + type(libref).__name__)
	return lib

sys = load('sys')

__docformat__ = 'reStructuredText'



##
# python.datetime - parts needed to use stdlib.datetime
##
import datetime

##
# stdlib.datetime representation of PostgreSQL 'infinity' and '-infinity'.
infinity_datetime = datetime.datetime(datetime.MAXYEAR, 12, 31, 23, 59, 59, 999999)
negative_infinity_datetime = datetime.datetime(datetime.MINYEAR, 1, 1, 0, 0, 0, 0)

infinity_date = datetime.date(datetime.MAXYEAR, 12, 31)
negative_infinity_date = datetime.date(datetime.MINYEAR, 1, 1)

class FixedOffset(datetime.tzinfo):
	def __init__(self, offset_in_seconds, tzname = None):
		self._tzname = tzname
		self._offset = offset_in_seconds
		self._offset_in_mins = offset_in_seconds // 60
		self._td_offset = datetime.timedelta(0, self._offset_in_mins * 60)
		self._dst = datetime.timedelta(0)

	def utcoffset(self, offset_from):
		return self._td_offset

	def tzname(self, dt):
		return self._tzname

	def dst(self, arg):
		return self._dst

	def __repr__(self):
		return "{path}.{name}({off}{tzname})".format(
			path = type(self).__module__,
			name = type(self).__name__,
			off = repr(self._td_offset.days * 24 * 60 * 60 + self._td_offset.seconds),
			tzname = (
				", tzname = {tzname!r}".format(tzname = self._tzname) \
				if self._tzname is not None else ""
			)
		)

UTC = FixedOffset(0, tzname = 'UTC')



##
# .exceptions - Exception hierarchy for PostgreSQL database ERRORs.
##
"""
PostgreSQL exceptions and warnings with associated state codes.

The primary entry points of this module is the `ErrorLookup` function and the
`WarningLookup` function. Given an SQL state code, they give back the most
appropriate Error or Warning subclass.

For more information on error codes see:
 http://www.postgresql.org/docs/current/static/errcodes-appendix.html

This module is executable via -m: python -m postgresql.exceptions.
It provides a convenient way to look up the exception object mapped to by the
given error code::

	$ python -m postgresql.exceptions XX000
	postgresql.exceptions.InternalError [XX000]

If the exact error code is not found, it will try to find the error class's
exception(The first two characters of the error code make up the class
identity)::

	$ python -m postgresql.exceptions XX400
	postgresql.exceptions.InternalError [XX000]

If that fails, it will return `postgresql.exceptions.Error`
"""
import sys
import os
from functools import partial
from operator import attrgetter
from .message import Message
from . import sys as pg_sys

PythonException = Exception
class Exception(Exception):
	"""
	Base PostgreSQL exception class.
	"""
	pass

class LoadError(Exception):
	"""
	Failed to load a library.
	"""

class Disconnection(Exception):
	"""
	Exception identifying errors that result in disconnection.
	"""

class Warning(Message):
	code = '01000'
	_e_label = property(attrgetter('__class__.__name__'))

class DriverWarning(Warning):
	code = '01-00'
	source = 'CLIENT'
class IgnoredClientParameterWarning(DriverWarning):
	'Warn the user of a valid, but ignored parameter.'
	code = '01-CP'
class TypeConversionWarning(DriverWarning):
	'Report a potential issue with a conversion.'
	code = '01-TP'

class DeprecationWarning(Warning):
	code = '01P01'
class DynamicResultSetsReturnedWarning(Warning):
	code = '0100C'
class ImplicitZeroBitPaddingWarning(Warning):
	code = '01008'
class NullValueEliminatedInSetFunctionWarning(Warning):
	code = '01003'
class PrivilegeNotGrantedWarning(Warning):
	code = '01007'
class PrivilegeNotRevokedWarning(Warning):
	code = '01006'
class StringDataRightTruncationWarning(Warning):
	code = '01004'

class NoDataWarning(Warning):
	code = '02000'
class NoMoreSetsReturned(NoDataWarning):
	code = '02001'

class Error(Message, Exception):
	"""
	A PostgreSQL Error.
	"""
	_e_label = 'ERROR'
	code = ''

	def __str__(self):
		"""
		Call .sys.errformat(self).
		"""
		return pg_sys.errformat(self)

	@property
	def fatal(self):
		f = self.details.get('severity')
		return None if f is None else f in ('PANIC', 'FATAL')

class DriverError(Error):
	"""
	Errors originating in the driver's implementation.
	"""
	source = 'CLIENT'
	code = '--000'
class AuthenticationMethodError(DriverError, Disconnection):
	"""
	Server requested an authentication method that is not supported by the
	driver.
	"""
	code = '--AUT'
class InsecurityError(DriverError, Disconnection):
	"""
	Error signifying a secure channel to a server cannot be established.
	"""
	code = '--SEC'
class ConnectTimeoutError(DriverError, Disconnection):
	"""
	Client was unable to esablish a connection in the given time.
	"""
	code = '--TOE'

class TypeIOError(DriverError):
	"""
	Driver failed to pack or unpack a tuple.
	"""
	code = '--TIO'
class ParameterError(TypeIOError):
	code = '--PIO'
class ColumnError(TypeIOError):
	code = '--CIO'
class CompositeError(TypeIOError):
	code = '--cIO'

class OperationError(DriverError):
	"""
	An invalid operation on an interface element.
	"""
	code = '--OPE'

class TransactionError(Error):
	pass

class SQLNotYetCompleteError(Error):
	code = '03000'

class ConnectionError(Error, Disconnection):
	code = '08000'
class ConnectionDoesNotExistError(ConnectionError):
	"""
	The connection is closed or was never connected.
	"""
	code = '08003'
class ConnectionFailureError(ConnectionError):
	"""
	Raised when a connection is dropped.
	"""
	code = '08006'

class ClientCannotConnectError(ConnectionError):
	"""
	Client was unable to establish a connection to the server.
	"""
	code = '08001'

class ConnectionRejectionError(ConnectionError):
	code = '08004'
class TransactionResolutionUnknownError(ConnectionError):
	code = '08007'
class ProtocolError(ConnectionError):
	code = '08P01'

class TriggeredActionError(Error):
	code = '09000'

class FeatureError(Error):
	"""
	"Unsupported feature.
	"""
	code = '0A000'

class TransactionInitiationError(TransactionError):
	code = '0B000'

class LocatorError(Error):
	code = '0F000'
class LocatorSpecificationError(LocatorError):
	code = '0F001'

class GrantorError(Error):
	code = '0L000'
class GrantorOperationError(GrantorError):
	code = '0LP01'

class RoleSpecificationError(Error):
	code = '0P000'

class CaseNotFoundError(Error):
	code = '20000'

class CardinalityError(Error):
	"""
	Wrong number of rows returned.
	"""
	code = '21000'

class TriggeredDataChangeViolation(Error):
	code = '27000'

class AuthenticationSpecificationError(Error, Disconnection):
	code = '28000'

class DPDSEError(Error):
	"""
	Dependent Privilege Descriptors Still Exist.
	"""
	code = '2B000'
class DPDSEObjectError(DPDSEError):
	code = '2BP01'

class SREError(Error):
	"""
	SQL Routine Exception.
	"""
	code = '2F000'
class FunctionExecutedNoReturnStatementError(SREError):
	code = '2F005'
class DataModificationProhibitedError(SREError):
	code = '2F002'
class StatementProhibitedError(SREError):
	code = '2F003'
class ReadingDataProhibitedError(SREError):
	code = '2F004'

class EREError(Error):
	"""
	External Routine Exception.
	"""
	code = '38000'
class ContainingSQLNotPermittedError(EREError):
	code = '38001'
class ModifyingSQLDataNotPermittedError(EREError):
	code = '38002'
class ProhibitedSQLStatementError(EREError):
	code = '38003'
class ReadingSQLDataNotPermittedError(EREError):
	code = '38004'

class ERIEError(Error):
	"""
	External Routine Invocation Exception.
	"""
	code = '39000'
class InvalidSQLState(ERIEError):
	code = '39001'
class NullValueNotAllowed(ERIEError):
	code = '39004'
class TriggerProtocolError(ERIEError):
	code = '39P01'
class SRFProtocolError(ERIEError):
	code = '39P02'

class TRError(TransactionError):
	"""
	Transaction Rollback.
	"""
	code = '40000'
class DeadlockError(TRError):
	code = '40P01'
class IntegrityConstraintViolationError(TRError):
	code = '40002'
class SerializationError(TRError):
	code = '40001'
class StatementCompletionUnknownError(TRError):
	code = '40003'


class ITSError(TransactionError):
	"""
	Invalid Transaction State.
	"""
	code = '25000'
class ActiveTransactionError(ITSError):
	code = '25001'
class BranchAlreadyActiveError(ITSError):
	code = '25002'
class BadAccessModeForBranchError(ITSError):
	code = '25003'
class BadIsolationForBranchError(ITSError):
	code = '25004'
class NoActiveTransactionForBranchError(ITSError):
	code = '25005'
class ReadOnlyTransactionError(ITSError):
	"""
	Occurs when an alteration occurs in a read-only transaction.
	"""
	code = '25006'
class SchemaAndDataStatementsError(ITSError):
	"""
	Mixed schema and data statements not allowed.
	"""
	code = '25007'
class InconsistentCursorIsolationError(ITSError):
	"""
	The held cursor requires the same isolation.
	"""
	code = '25008'

class NoActiveTransactionError(ITSError):
	code = '25P01'
class InFailedTransactionError(ITSError):
	"""
	Occurs when an action occurs in a failed transaction.
	"""
	code = '25P02'


class SavepointError(TransactionError):
	"""
	Classification error designating errors that relate to savepoints.
	"""
	code = '3B000'
class InvalidSavepointSpecificationError(SavepointError):
	code = '3B001'

class TransactionTerminationError(TransactionError):
	code = '2D000'

class IRError(Error):
	"""
	Insufficient Resource Error.
	"""
	code = '53000'
class MemoryError(IRError, MemoryError):
	code = '53200'
class DiskFullError(IRError):
	code = '53100'
class TooManyConnectionsError(IRError):
	code = '53300'

class PLEError(OverflowError):
	"""
	Program Limit Exceeded
	"""
	code = '54000'
class ComplexityOverflowError(PLEError):
	code = '54001'
class ColumnOverflowError(PLEError):
	code = '54011'
class ArgumentOverflowError(PLEError):
	code = '54023'

class ONIPSError(Error):
	"""
	Object Not In Prerequisite State.
	"""
	code = '55000'
class ObjectInUseError(ONIPSError):
	code = '55006'
class ImmutableRuntimeParameterError(ONIPSError):
	code = '55P02'
class UnavailableLockError(ONIPSError):
	code = '55P03'


class SEARVError(Error):
	"""
	Syntax Error or Access Rule Violation.
	"""
	code = '42000'

class SEARVNameError(SEARVError):
	code = '42602'
class NameTooLongError(SEARVError):
	code = '42622'
class ReservedNameError(SEARVError):
	code = '42939'

class ForeignKeyCreationError(SEARVError):
	code = '42830'

class InsufficientPrivilegeError(SEARVError):
	code = '42501'
class GroupingError(SEARVError):
	code = '42803'

class RecursionError(SEARVError):
	code = '42P19'
class WindowError(SEARVError):
	code = '42P20'

class SyntaxError(SEARVError):
	code = '42601'

class TypeError(SEARVError):
	pass
class CoercionError(TypeError):
	code = '42846'
class TypeMismatchError(TypeError):
	code = '42804'
class IndeterminateTypeError(TypeError):
	code = '42P18'
class WrongObjectTypeError(TypeError):
	code = '42809'

class UndefinedError(SEARVError):
	pass
class UndefinedColumnError(UndefinedError):
	code = '42703'
class UndefinedFunctionError(UndefinedError):
	code = '42883'
class UndefinedTableError(UndefinedError):
	code = '42P01'
class UndefinedParameterError(UndefinedError):
	code = '42P02'
class UndefinedObjectError(UndefinedError):
	code = '42704'

class DuplicateError(SEARVError):
	pass
class DuplicateColumnError(DuplicateError):
	code = '42701'
class DuplicateCursorError(DuplicateError):
	code = '42P03'
class DuplicateDatabaseError(DuplicateError):
	code = '42P04'
class DuplicateFunctionError(DuplicateError):
	code = '42723'
class DuplicatePreparedStatementError(DuplicateError):
	code = '42P05'
class DuplicateSchemaError(DuplicateError):
	code = '42P06'
class DuplicateTableError(DuplicateError):
	code = '42P07'
class DuplicateAliasError(DuplicateError):
	code = '42712'
class DuplicateObjectError(DuplicateError):
	code = '42710'

class AmbiguityError(SEARVError):
	pass
class AmbiguousColumnError(AmbiguityError):
	code = '42702'
class AmbiguousFunctionError(AmbiguityError):
	code = '42725'
class AmbiguousParameterError(AmbiguityError):
	code = '42P08'
class AmbiguousAliasError(AmbiguityError):
	code = '42P09'

class ColumnReferenceError(SEARVError):
	code = '42P10'

class DefinitionError(SEARVError):
	pass
class ColumnDefinitionError(DefinitionError):
	code = '42611'
class CursorDefinitionError(DefinitionError):
	code = '42P11'
class DatabaseDefinitionError(DefinitionError):
	code = '42P12'
class FunctionDefinitionError(DefinitionError):
	code = '42P13'
class PreparedStatementDefinitionError(DefinitionError):
	code = '42P14'
class SchemaDefinitionError(DefinitionError):
	code = '42P15'
class TableDefinitionError(DefinitionError):
	code = '42P16'
class ObjectDefinitionError(DefinitionError):
	code = '42P17'


class CursorStateError(Error):
	code = '24000'

class WithCheckOptionError(Error):
	code = '44000'

class NameError(Error):
	pass
class CatalogNameError(NameError):
	code = '3D000'
class CursorNameError(NameError):
	code = '34000'
class StatementNameError(NameError):
	code = '26000'
class SchemaNameError(NameError):
	code = '3F000'

class ICVError(Error):
	"""
	Integrity Contraint Violation.
	"""
	code = '23000'
class RestrictError(ICVError):
	code = '23001'
class NotNullError(ICVError):
	code = '23502'
class ForeignKeyError(ICVError):
	code = '23503'
class UniqueError(ICVError):
	code = '23505'
class CheckError(ICVError):
	code = '23514'


class DataError(Error):
	code = '22000'

class StringRightTruncationError(DataError):
	code = '22001'
class StringDataLengthError(DataError):
	code = '22026'
class ZeroLengthString(DataError):
	code = '2200F'

class EncodingError(DataError):
	code = '22021'
class ArrayElementError(DataError):
	code = '2202E'
class SpecificTypeMismatch(DataError):
	code = '2200G'

class NullValueNotAllowedError(DataError):
	code = '22004'
class NullValueNoIndicatorParameter(DataError):
	code = '22002'

class ZeroDivisionError(DataError):
	code = '22012'
class FloatingPointError(DataError):
	code = '22P01'
class AssignmentError(DataError):
	code = '22005'
class IndicatorOverflowError(DataError):
	code = '22022'
class BadCopyError(DataError):
	code = '22P04'

class TextRepresentationError(DataError):
	code = '22P02'
class BinaryRepresentationError(DataError):
	code = '22P03'
class UntranslatableCharacterError(DataError):
	code = '22P05'
class NonstandardUseOfEscapeCharacterError(DataError):
	code = '22P06'

class NotXMLError(DataError):
	code = '2200L'
class XMLDocumentError(DataError):
	code = '2200M'
class XMLContentError(DataError):
	code = '2200N'
class XMLCommentError(DataError):
	code = '2200S'
class XMLProcessingInstructionError(DataError):
	code = '2200T'

class DateTimeFormatError(DataError):
	code = '22007'
class TimeZoneDisplacementValueError(DataError):
	code = '22009'
class DateTimeFieldOverflowError(DataError):
	code = '22008'
class IntervalFieldOverflowError(DataError):
	code = '22015'

class LogArgumentError(DataError):
	code = '2201E'
class PowerFunctionArgumentError(DataError):
	code = '2201F'
class WidthBucketFunctionArgumentError(DataError):
	code = '2201G'
class CastCharacterValueError(DataError):
	code = '22018'

class EscapeCharacterError(DataError):
	code = '22019'
class EscapeOctetError(DataError):
	code = '2200D'
class EscapeSequenceError(DataError):
	code = '22025'
class EscapeCharacterConflictError(DataError):
	code = '2200B'
class EscapeCharacterError(DataError):
	"""
	Invalid escape character.
	"""
	code = '2200C'

class SubstringError(DataError):
	code = '22011'
class TrimError(DataError):
	code = '22027'
class IndicatorParameterValueError(DataError):
	code = '22010'

class LimitValueError(DataError):
	code = '2201W'
	pg_code = '22020'
class OffsetValueError(DataError):
	code = '2201X'

class ParameterValueError(DataError):
	code = '22023'
class RegularExpressionError(DataError):
	code = '2201B'
class NumericRangeError(DataError):
	code = '22003'
class UnterminatedCStringError(DataError):
	code = '22024'


class InternalError(Error):
	code = 'XX000'
class DataCorruptedError(InternalError):
	code = 'XX001'
class IndexCorruptedError(InternalError):
	code = 'XX002'

class SIOError(Error):
	"""
	System I/O.
	"""
	code = '58000'
class UndefinedFileError(SIOError):
	code = '58P01'
class DuplicateFileError(SIOError):
	code = '58P02'

class CFError(Error):
	"""
	Configuration File Error.
	"""
	code = 'F0000'
class LockFileExistsError(CFError):
	code = 'F0001'

class OIError(Error):
	"""
	Operator Intervention.
	"""
	code = '57000'
class QueryCanceledError(OIError):
	code = '57014'
class AdminShutdownError(OIError, Disconnection):
	code = '57P01'
class CrashShutdownError(OIError, Disconnection):
	code = '57P02'
class ServerNotReadyError(OIError, Disconnection):
	"""
	Thrown when a connection is established to a server that is still starting up.
	"""
	code = '57P03'

class PLPGSQLError(Error):
	"""
	Error raised by a PL/PgSQL procedural function.
	"""
	code = 'P0000'
class PLPGSQLRaiseError(PLPGSQLError):
	"""
	Error raised by a PL/PgSQL RAISE statement.
	"""
	code = 'P0001'
class PLPGSQLNoDataFoundError(PLPGSQLError):
	code = 'P0002'
class PLPGSQLTooManyRowsError(PLPGSQLError):
	code = 'P0003'


# Setup mapping to provide code based exception lookup.
code_to_error = {}
code_to_warning = {}
def map_errors_and_warnings(
	objs,
	error_container = code_to_error,
	warning_container = code_to_warning,
):
	"""
	Construct the code-to-error and code-to-warning associations.
	"""
	for obj in objs:
		if not issubclass(type(obj), (type(Warning), type(Error))):
			# It's not object of interest.
			continue
		code = getattr(obj, 'code', None)
		if code is None:
			# It has no code attribute, or the code was set to None.
			# If it's code is None, we don't map it as it's a "container".
			continue

		if issubclass(obj, Error):
			base = Error
			container = error_container
		elif issubclass(obj, Warning):
			base = Warning
			container = warning_container
		else:
			continue

		cur_obj = container.get(code)
		if cur_obj is None or issubclass(cur_obj, obj):
			# There is no object yet, or the object at the code
			# is not the most general class.
			# The latter condition comes into play when
			# there are sub-Class types that share the Class code
			# with the most general type. (See TypeError)
			container[code] = obj
			if hasattr(obj, 'pg_code'):
				# If there's a PostgreSQL version of the code,
				# map it as well for older servers.
				container[obj.pg_code] = obj

def code_lookup(
	default,
	container,
	code
):
	obj = container.get(code)
	if obj is None:
		obj = container.get(code[:2] + "000", default)
	return obj

map_errors_and_warnings(sys.modules[__name__].__dict__.values())
ErrorLookup = partial(code_lookup, Error, code_to_error)
WarningLookup = partial(code_lookup, Warning, code_to_warning)

if __name__ == '__main__':
	for x in sys.argv[1:]:
		if x.startswith('01'):
			e = WarningLookup(x)
		else:
			e = ErrorLookup(x)
		sys.stdout.write('postgresql.exceptions.%s [%s]%s%s' %(
				e.__name__, e.code, os.linesep, (
					e.__doc__ is not None and os.linesep.join([
						'  ' + x for x in (e.__doc__).split('\n')
					]) + os.linesep or ''
				)
			)
		)



"""
Project information.
"""

name = 'py-opengauss'
identity = 'http://github.com/vimiix/py-opengauss'

meaculpa = 'Python+openGauss'
abstract = 'Driver and tools library for openGauss'

version_info = (1, 3, 4)  # dev based on py-postgresql version 1.3.0
version = '.'.join(map(str, version_info))



##
# .pgpassfile - parse and lookup passwords in a pgpassfile
##
"""
Parse pgpass files and subsequently lookup a password.
"""
import os.path

def split(line, len = len):
	line = line.strip()
	if not line:
		return None
	r = []
	continuation = False
	for x in line.split(':'):
		if continuation:
			# The colon was preceded by a backslash, it's part
			# of the last field. Substitute the trailing backslash
			# with the colon and append the next value.
			r[-1] = r[-1][:-1] + ':' + x.replace('\\\\', '\\')
			continuation = False
		else:
			# Even number of backslashes preceded the split.
			# Normal field.
			r.append(x.replace('\\\\', '\\'))
		# Determine if the next field is a continuation of this one.
		if (len(x) - len(x.rstrip('\\'))) % 2 == 1:
			continuation = True
	if len(r) != 5:
		# Too few or too many fields.
		return None
	return r

def parse(data):
	"""
	Produce a list of [(word, (host,port,dbname,user))] from a pgpass file object.
	"""
	return [
		(x[-1], x[0:4]) for x in [split(line) for line in data] if x
	]

def lookup_password(words, uhpd):
	"""
	lookup_password(words, (user, host, port, database)) -> password

	Where 'words' is the output from pgpass.parse()
	"""
	user, host, port, database = uhpd
	for word, (w_host, w_port, w_database, w_user) in words:
		if (w_user == '*' or w_user == user) and \
			(w_host == '*' or w_host == host) and \
			(w_port == '*' or w_port == port) and \
		(w_database == '*' or w_database == database):
			return word

def lookup_password_file(path, t):
	"""
	Like lookup_password, but takes a file path.
	"""
	with open(path) as f:
		return lookup_password(parse(f), t)

def lookup_pgpass(d, passfile, exists = os.path.exists):
	# If the password file exists, lookup the password
	# using the config's criteria.
	if exists(passfile):
		return lookup_password_file(passfile, (
			str(d['user']), str(d['host']), str(d['port']),
			str(d.get('database', d['user']))
		))



from ...types import JSONBOID


def jsonb_pack(x, typeio):
	jsonb = typeio.encode(x)
	return b'\x01' + jsonb


def jsonb_unpack(x, typeio):
	if x[0] != 1:
		raise ValueError('unexpected JSONB format version: {!r}'.format(x[0]))
	return typeio.decode(x[1:])


def _jsonb_io_factory(oid, typeio):
	_pack = lambda x: jsonb_pack(x, typeio)
	_unpack = lambda x: jsonb_unpack(x, typeio)

	return (_pack, _unpack, str)


oid_to_io = {
	JSONBOID: _jsonb_io_factory
}



##
# .test.test_types - test type representations and I/O
##
import unittest
import struct
from ..python.functools import process_tuple
from .. import types as pg_types
from ..types.io import lib as typlib
from ..types.io import builtins
from ..types.io.contrib_hstore import hstore_factory
from ..types import Array

class fake_typio(object):
	@staticmethod
	def encode(x):
		return x.encode('utf-8')
	@staticmethod
	def decode(x):
		return x.decode('utf-8')
hstore_pack, hstore_unpack = hstore_factory(0, fake_typio)

# this must pack to that, and
# that must unpack to this
expectation_samples = {
	('bool', lambda x: builtins.bool_pack(x), lambda x: builtins.bool_unpack(x)) : [
		(True, b'\x01'),
		(False, b'\x00'),
	],

	('int2', builtins.int2_pack, builtins.int2_unpack) : [
		(0, b'\x00\x00'),
		(1, b'\x00\x01'),
		(2, b'\x00\x02'),
		(0x0f, b'\x00\x0f'),
		(0xf00, b'\x0f\x00'),
		(0x7fff, b'\x7f\xff'),
		(-0x8000, b'\x80\x00'),
		(-1, b'\xff\xff'),
		(-2, b'\xff\xfe'),
		(-3, b'\xff\xfd'),
	],

	('int4', builtins.int4_pack, builtins.int4_unpack) : [
		(0, b'\x00\x00\x00\x00'),
		(1, b'\x00\x00\x00\x01'),
		(2, b'\x00\x00\x00\x02'),
		(0x0f, b'\x00\x00\x00\x0f'),
		(0x7fff, b'\x00\x00\x7f\xff'),
		(-0x8000, b'\xff\xff\x80\x00'),
		(0x7fffffff, b'\x7f\xff\xff\xff'),
		(-0x80000000, b'\x80\x00\x00\x00'),
		(-1, b'\xff\xff\xff\xff'),
		(-2, b'\xff\xff\xff\xfe'),
		(-3, b'\xff\xff\xff\xfd'),
	],

	('int8', builtins.int8_pack, builtins.int8_unpack) : [
		(0, b'\x00\x00\x00\x00\x00\x00\x00\x00'),
		(1, b'\x00\x00\x00\x00\x00\x00\x00\x01'),
		(2, b'\x00\x00\x00\x00\x00\x00\x00\x02'),
		(0x0f, b'\x00\x00\x00\x00\x00\x00\x00\x0f'),
		(0x7fffffff, b'\x00\x00\x00\x00\x7f\xff\xff\xff'),
		(0x80000000, b'\x00\x00\x00\x00\x80\x00\x00\x00'),
		(-0x80000000, b'\xff\xff\xff\xff\x80\x00\x00\x00'),
		(-1, b'\xff\xff\xff\xff\xff\xff\xff\xff'),
		(-2, b'\xff\xff\xff\xff\xff\xff\xff\xfe'),
		(-3, b'\xff\xff\xff\xff\xff\xff\xff\xfd'),
	],

	('numeric', typlib.numeric_pack, typlib.numeric_unpack) : [
		(((0,0,0,0),[]), b'\x00'*2*4),
		(((0,0,0,0),[1]), b'\x00'*2*4 + b'\x00\x01'),
		(((1,0,0,0),[1]), b'\x00\x01' + b'\x00'*2*3 + b'\x00\x01'),
		(((1,1,1,1),[1]), b'\x00\x01'*4 + b'\x00\x01'),
		(((1,1,1,1),[1,2]), b'\x00\x01'*4 + b'\x00\x01\x00\x02'),
		(((1,1,1,1),[1,2,3]), b'\x00\x01'*4 + b'\x00\x01\x00\x02\x00\x03'),
	],

	('varbit', typlib.varbit_pack, typlib.varbit_unpack) : [
		((0, b'\x00'), b'\x00\x00\x00\x00\x00'),
		((1, b'\x01'), b'\x00\x00\x00\x01\x01'),
		((1, b'\x00'), b'\x00\x00\x00\x01\x00'),
		((2, b'\x00'), b'\x00\x00\x00\x02\x00'),
		((3, b'\x00'), b'\x00\x00\x00\x03\x00'),
		((9, b'\x00\x00'), b'\x00\x00\x00\x09\x00\x00'),
		# More data than necessary, we allow this.
		# Let the user do the necessary check if the cost is worth the benefit.
		((9, b'\x00\x00\x00'), b'\x00\x00\x00\x09\x00\x00\x00'),
	],

	# idk why
	('bytea', builtins.bytea_pack, builtins.bytea_unpack) : [
		(b'foo', b'foo'),
		(b'bar', b'bar'),
		(b'\x00', b'\x00'),
		(b'\x01', b'\x01'),
	],

	('char', builtins.char_pack, builtins.char_unpack) : [
		(b'a', b'a'),
		(b'b', b'b'),
		(b'\x00', b'\x00'),
	],

	('point', typlib.point_pack, typlib.point_unpack) : [
		((1.0, 1.0), b'?\xf0\x00\x00\x00\x00\x00\x00?\xf0\x00\x00\x00\x00\x00\x00'),
		((2.0, 2.0), b'@\x00\x00\x00\x00\x00\x00\x00@\x00\x00\x00\x00\x00\x00\x00'),
		((-1.0, -1.0),
			b'\xbf\xf0\x00\x00\x00\x00\x00\x00\xbf\xf0\x00\x00\x00\x00\x00\x00'),
	],

	('circle', typlib.circle_pack, typlib.circle_unpack) : [
		((1.0, 1.0, 1.0),
			b'?\xf0\x00\x00\x00\x00\x00\x00?\xf0\x00\x00' \
			b'\x00\x00\x00\x00?\xf0\x00\x00\x00\x00\x00\x00'),
		((2.0, 2.0, 2.0),
			b'@\x00\x00\x00\x00\x00\x00\x00@\x00\x00\x00' \
			b'\x00\x00\x00\x00@\x00\x00\x00\x00\x00\x00\x00'),
	],

	('record', typlib.record_pack, typlib.record_unpack) : [
		([], b'\x00\x00\x00\x00'),
		([(0,b'foo')], b'\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x03foo'),
		([(0,None)], b'\x00\x00\x00\x01\x00\x00\x00\x00\xff\xff\xff\xff'),
		([(15,None)], b'\x00\x00\x00\x01\x00\x00\x00\x0f\xff\xff\xff\xff'),
		([(0xffffffff,None)], b'\x00\x00\x00\x01\xff\xff\xff\xff\xff\xff\xff\xff'),
		([(0,None), (1,b'some')],
		 b'\x00\x00\x00\x02\x00\x00\x00\x00\xff\xff\xff\xff' \
		 b'\x00\x00\x00\x01\x00\x00\x00\x04some'),
	],

	('array', typlib.array_pack, typlib.array_unpack) : [
		([0, 0xf, (1,), (0,), (b'foo',)],
			b'\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x0f\x00\x00\x00\x01' \
			b'\x00\x00\x00\x00\x00\x00\x00\x03foo'
		),
		([0, 0xf, (1,), (0,), (None,)],
			b'\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x0f\x00\x00\x00\x01' \
			b'\x00\x00\x00\x00\xff\xff\xff\xff'
		)
	],

	('hstore', hstore_pack, hstore_unpack) : [
		({}, b'\x00\x00\x00\x00'),
		({'b' : None}, b'\x00\x00\x00\x01\x00\x00\x00\x01b\xff\xff\xff\xff'),
		({'b' : 'k'}, b'\x00\x00\x00\x01\x00\x00\x00\x01b\x00\x00\x00\x01k'),
		({'foo' : 'bar'}, b'\x00\x00\x00\x01\x00\x00\x00\x03foo\x00\x00\x00\x03bar'),
		({'foo' : None}, b'\x00\x00\x00\x01\x00\x00\x00\x03foo\xff\xff\xff\xff'),
	],
}
expectation_samples[('box', typlib.box_pack, typlib.box_unpack)] = \
	expectation_samples[('lseg', typlib.lseg_pack, typlib.lseg_unpack)] = [
		((1.0, 1.0, 1.0, 1.0),
			b'?\xf0\x00\x00\x00\x00\x00\x00?\xf0' \
			b'\x00\x00\x00\x00\x00\x00?\xf0\x00\x00' \
			b'\x00\x00\x00\x00?\xf0\x00\x00\x00\x00\x00\x00'),
		((2.0, 2.0, 1.0, 1.0),
			b'@\x00\x00\x00\x00\x00\x00\x00@\x00\x00' \
			b'\x00\x00\x00\x00\x00?\xf0\x00\x00\x00\x00' \
			b'\x00\x00?\xf0\x00\x00\x00\x00\x00\x00'),
		((-1.0, -1.0, 1.0, 1.0),
			b'\xbf\xf0\x00\x00\x00\x00\x00\x00\xbf\xf0' \
			b'\x00\x00\x00\x00\x00\x00?\xf0\x00\x00\x00' \
			b'\x00\x00\x00?\xf0\x00\x00\x00\x00\x00\x00'),
	]

expectation_samples[('oid', typlib.oid_pack, typlib.oid_unpack)] = \
	expectation_samples[('cid', typlib.cid_pack, typlib.cid_unpack)] = \
	expectation_samples[('xid', typlib.xid_pack, typlib.xid_unpack)] = [
		(0, b'\x00\x00\x00\x00'),
		(1, b'\x00\x00\x00\x01'),
		(2, b'\x00\x00\x00\x02'),
		(0xf, b'\x00\x00\x00\x0f'),
		(0xffffffff, b'\xff\xff\xff\xff'),
		(0x7fffffff, b'\x7f\xff\xff\xff'),
	]

# this must pack and then unpack back into this
consistency_samples = {
	('bool', lambda x: builtins.bool_pack(x), lambda x: builtins.bool_unpack(x)) : [True, False],

	('record', typlib.record_pack, typlib.record_unpack) : [
		[],
		[(0,b'foo')],
		[(0,None)],
		[(15,None)],
		[(0xffffffff,None)],
		[(0,None), (1,b'some')],
		[(0,None), (1,b'some'), (0xffff, b"something_else\x00")],
		[(0,None), (1,b"s\x00me"), (0xffff, b"\x00something_else\x00")],
	],

	('array', typlib.array_pack, typlib.array_unpack) : [
		[0, 0xf, (), (), ()],
		[0, 0xf, (0,), (0,), ()],
		[0, 0xf, (1,), (0,), (b'foo',)],
		[0, 0xf, (1,), (0,), (None,)],
		[0, 0xf, (2,), (0,), (None,None)],
		[0, 0xf, (2,), (0,), (b'foo',None)],
		[0, 0xff, (2,), (0,), (None,b'foo',)],
		[0, 0xffffffff, (3,), (0,), (None,b'foo',None)],
		[1, 0xffffffff, (3,), (0,), (None,b'foo',None)],
		[1, 0xffffffff, (3, 1), (0, 0), (None,b'foo',None)],
		[1, 0xffffffff, (3, 2), (0, 0), (None,b'one',b'foo',b'two',None,b'three')],
	],

	# Just some random data; it's just an integer, so nothing fancy.
	('date', typlib.date_pack, typlib.date_unpack) : [
		123,
		321,
		0x7FFFFFF,
		-0x8000000,
	],

	('point', typlib.point_pack, typlib.point_unpack) : [
		(0, 0),
		(2, 2),
		(-1, -1),
		(-1.5, -1.2),
		(1.5, 1.2),
	],

	('circle', typlib.circle_pack, typlib.circle_unpack) : [
		(0, 0, 0),
		(2, 2, 2),
		(-1, -1, -1),
		(-1.5, -1.2, -1.8),
	],

	('tid', typlib.tid_pack, typlib.tid_unpack) : [
		(0, 0),
		(1, 1),
		(0xffffffff, 0xffff),
		(0, 0xffff),
		(0xffffffff, 0),
		(0xffffffff // 2, 0xffff // 2),
	],
}
__ = {
	('cidr', typlib.net_pack, typlib.net_unpack) : [
		(0, 0, b"\x00\x00\x00\x00"),
		(2, 0, b"\x00" * 4),
		(2, 0, b"\xFF" * 4),
		(2, 32, b"\xFF" * 4),
		(3, 0, b"\x00\x00" * 16),
	],

	('inet', typlib.net_pack, typlib.net_unpack) : [
		(2, 32, b"\x00\x00\x00\x00"),
		(2, 16, b"\x7f\x00\x00\x01"),
		(2, 8, b"\xff\x00\xff\x01"),
		(3, 128, b"\x7f\x00" * 16),
		(3, 64, b"\xff\xff" * 16),
		(3, 32, b"\x00\x00" * 16),
	],
}

consistency_samples[('time', typlib.time_pack, typlib.time_unpack)] = \
consistency_samples[('time64', typlib.time64_pack, typlib.time64_unpack)] = [
	(0, 0),
	(123, 123),
	(0xFFFFFFFF, 999999),
]

# months, days, (seconds, microseconds)
consistency_samples[('interval', typlib.interval_pack, typlib.interval_unpack)] = [
	(0, 0, (0, 0)),
	(1, 0, (0, 0)),
	(0, 1, (0, 0)),
	(1, 1, (0, 0)),
	(0, 0, (0, 10000)),
	(0, 0, (1, 0)),
	(0, 0, (1, 10000)),
	(1, 1, (1, 10000)),
	(100, 50, (1423, 29313))
]

consistency_samples[('timetz', typlib.timetz_pack, typlib.timetz_unpack)] = \
consistency_samples[('timetz', typlib.timetz64_pack, typlib.timetz64_unpack)] = \
	[
		((0, 0), 0),
		((123, 123), 123),
		((0xFFFFFFFF, 999999), -123),
	]

consistency_samples[('oid', typlib.oid_pack, typlib.oid_unpack)] = \
	consistency_samples[('cid', typlib.cid_pack, typlib.cid_unpack)] = \
	consistency_samples[('xid', typlib.xid_pack, typlib.xid_unpack)] = [
	0, 0xffffffff, 0xffffffff // 2, 123, 321, 1, 2, 3
]

consistency_samples[('lseg', typlib.lseg_pack, typlib.lseg_unpack)] = \
	consistency_samples[('box', typlib.box_pack, typlib.box_unpack)] = [
	(1,2,3,4),
	(4,3,2,1),
	(0,0,0,0),
	(-1,-1,-1,-1),
	(-1.2,-1.5,-2.0,4.0)
]

consistency_samples[('path', typlib.path_pack, typlib.path_unpack)] = \
	consistency_samples[('polygon', typlib.polygon_pack, typlib.polygon_unpack)] = [
	(1,2,3,4),
	(4,3,2,1),
	(0,0,0,0),
	(-1,-1,-1,-1),
	(-1.2,-1.5,-2.0,4.0),
]

from types import GeneratorType
def resolve(ob):
	'make sure generators get "tuplified"'
	if type(ob) not in (list, tuple, GeneratorType):
		return ob
	return [resolve(x) for x in ob]

def testExpectIO(self, samples):
	for id, sample in samples.items():
		name, pack, unpack = id

		for (sample_unpacked, sample_packed) in sample:
			pack_trial = pack(sample_unpacked)
			self.assertTrue(
				pack_trial == sample_packed,
				"%s sample: unpacked sample, %r, did not match " \
				"%r when packed, rather, %r" %(
					name, sample_unpacked,
					sample_packed, pack_trial
				)
			)

			sample_unpacked = resolve(sample_unpacked)
			unpack_trial = resolve(unpack(sample_packed))
			self.assertTrue(
				unpack_trial == sample_unpacked,
				"%s sample: packed sample, %r, did not match " \
				"%r when unpacked, rather, %r" %(
					name, sample_packed,
					sample_unpacked, unpack_trial
				)
			)

class test_io(unittest.TestCase):
	def test_process_tuple(self):
		def funpass(cause, procs, tup, col):
			pass
		self.assertEqual(tuple(process_tuple((),(), funpass)), ())
		self.assertEqual(tuple(process_tuple((int,),("100",), funpass)), (100,))
		self.assertEqual(tuple(process_tuple((int,int),("100","200"), funpass)), (100,200))
		self.assertEqual(tuple(process_tuple((int,int),(None,"200"), funpass)), (None,200))
		self.assertEqual(tuple(process_tuple((int,int,int),(None,None,"200"), funpass)), (None,None,200))
		# The exception handler must raise.
		self.assertRaises(RuntimeError, process_tuple, (int,), ("foo",), funpass)

		class ThisError(Exception):
			pass
		data = []
		def funraise(cause, procs, tup, col):
			data.append((procs, tup, col))
			raise ThisError from cause
		self.assertRaises(ThisError, process_tuple, (int,), ("foo",), funraise)
		self.assertEqual(data[0], ((int,), ("foo",), 0))
		del data[0]
		self.assertRaises(ThisError, process_tuple, (int,int), ("100","bar"), funraise)
		self.assertEqual(data[0], ((int,int), ("100","bar"), 1))

	def testExpectations(self):
		'IO tests where the pre-made expected serialized form is compared'
		testExpectIO(self, expectation_samples)

	def testConsistency(self):
		'IO tests where the unpacked source is compared to re-unpacked result'
		for id, sample in consistency_samples.items():
			name, pack, unpack = id
			if pack is not None:
				for x in sample:
					packed = pack(x)
					unpacked = resolve(unpack(packed))
					x = resolve(x)
					self.assertTrue(x == unpacked,
						"inconsistency with %s, %r -> %r -> %r" %(
							name, x, packed, unpacked
						)
					)

	##
	# Further hstore tests.
	def test_hstore(self):
		# Can't do some tests with the consistency checks
		# because we are not using ordered dictionaries.
		self.assertRaises((ValueError, struct.error), hstore_unpack, b'\x00\x00\x00\x00foo')
		self.assertRaises(ValueError, hstore_unpack, b'\x00\x00\x00\x01')
		self.assertRaises(ValueError, hstore_unpack, b'\x00\x00\x00\x02\x00\x00\x00\x01G\x00\x00\x00\x01G')
		sample = [
			([('foo','bar'),('k',None),('zero','heroes')],
				b'\x00\x00\x00\x03\x00\x00\x00\x03foo' + \
				b'\x00\x00\x00\x03bar\x00\x00\x00\x01k\xFF\xFF\xFF\xFF' + \
				b'\x00\x00\x00\x04zero\x00\x00\x00\x06heroes'),
			([('foo',None),('k',None),('zero',None)],
				b'\x00\x00\x00\x03\x00\x00\x00\x03foo' + \
				b'\xff\xff\xff\xff\x00\x00\x00\x01k\xFF\xFF\xFF\xFF' + \
				b'\x00\x00\x00\x04zero\xFF\xFF\xFF\xFF'),
			([], b'\x00\x00\x00\x00'),
		]
		for x in sample:
			src, serialized = x
			self.assertEqual(hstore_pack(src), serialized)
			self.assertEqual(hstore_unpack(serialized), dict(src))

# Make some slices; used by testSlicing
slice_samples = [
	slice(0, None, x+1) for x in range(10)
] + [
	slice(x, None, 1) for x in range(10)
] + [
	slice(None, x, 1) for x in range(10)
] + [
	slice(None, -x, 70) for x in range(10)
] + [
	slice(x+1, x, -1) for x in range(10)
] + [
	slice(x+4, x, -2) for x in range(10)
]

class test_Array(unittest.TestCase):
	def emptyArray(self, a):
		self.assertEqual(len(a), 0)
		self.assertEqual(list(a.elements()), [])
		self.assertEqual(a.dimensions, ())
		self.assertEqual(a.lowerbounds, ())
		self.assertEqual(a.upperbounds, ())
		self.assertRaises(IndexError, a.__getitem__, 0)

	def testArrayInstantiation(self):
		a = Array([])
		self.emptyArray(a)
		# exercise default upper/lower
		a = Array((1,2,3,))
		self.assertEqual((a[0],a[1],a[2]), (1,2,3,))
		# Python interface, Python semantics.
		self.assertRaises(IndexError, a.__getitem__, 3)
		self.assertEqual(a.dimensions, (3,))
		self.assertEqual(a.lowerbounds, (1,))
		self.assertEqual(a.upperbounds, (3,))

	def testNestedArrayInstantiation(self):
		a = Array(([1,2],[3,4]))
		# Python interface, Python semantics.
		self.assertRaises(IndexError, a.__getitem__, 3)
		self.assertEqual(a.dimensions, (2,2,))
		self.assertEqual(a.lowerbounds, (1,1))
		self.assertEqual(a.upperbounds, (2,2))
		self.assertEqual(list(a.elements()), [1,2,3,4])
		self.assertEqual(list(a),
			[
				Array([1, 2]),
				Array([3, 4]),
			]
		)

		a = Array(([[1],[2]],[[3],[4]]))
		self.assertRaises(IndexError, a.__getitem__, 3)
		self.assertEqual(a.dimensions, (2,2,1))
		self.assertEqual(a.lowerbounds, (1,1,1))
		self.assertEqual(a.upperbounds, (2,2,1))
		self.assertEqual(list(a),
			[
				Array([[1], [2]]),
				Array([[3], [4]]),
			]
		)

		self.assertRaises(ValueError, Array, [
			[1], [2,3]
		])
		self.assertRaises(ValueError, Array, [
			[1], []
		])
		self.assertRaises(ValueError, Array, [
			[[1]],
			[[],2]
		])
		self.assertRaises(ValueError, Array, [
			[[[[[1,2,3]]]]],
			[[[[[1,2,3]]]]],
			[[[[[1,2,3]]]]],
			[[[[[2,2]]]]],
		])

	def testSlicing(self):
		elements = [1,2,3,4,5,6,7,8]
		d1 = Array([1,2,3,4,5,6,7,8])
		for x in slice_samples:
			self.assertEqual(
				d1[x], Array(elements[x])
			)
		elements = [[1,2],[3,4],[5,6],[7,8]]
		d2 = Array(elements)
		for x in slice_samples:
			self.assertEqual(
				d2[x], Array(elements[x])
			)
		elements = [
			[[[1,2],[3,4]]],
			[[[5,6],[791,8]]],
			[[[1,2],[333,4]]],
			[[[1,2],[3,4]]],
			[[[5,10],[7,8]]],
			[[[0,6],[7,8]]],
			[[[1,2],[3,4]]],
			[[[5,6],[7,8]]],
		]
		d3 = Array(elements)
		for x in slice_samples:
			self.assertEqual(
				d3[x], Array(elements[x])
			)

	def testFromElements(self):
		a = Array.from_elements(())
		self.emptyArray(a)

		# exercise default upper/lower
		a = Array.from_elements((1,2,3,))
		self.assertEqual((a[0],a[1],a[2]), (1,2,3,))
		# Python interface, Python semantics.
		self.assertRaises(IndexError, a.__getitem__, 3)
		self.assertEqual(a.dimensions, (3,))
		self.assertEqual(a.lowerbounds, (1,))
		self.assertEqual(a.upperbounds, (3,))

		# exercise default upper/lower
		a = Array.from_elements([3,2,1], lowerbounds = (2,), upperbounds = (4,))
		self.assertEqual(a.dimensions, (3,))
		self.assertEqual(a.lowerbounds, (2,))
		self.assertEqual(a.upperbounds, (4,))

	def testEmptyDimension(self):
		self.assertRaises(ValueError,
			Array, [[]]
		)
		self.assertRaises(ValueError,
			Array, [[2],[]]
		)
		self.assertRaises(ValueError,
			Array, [[],[],[]]
		)
		self.assertRaises(ValueError,
			Array, [[2],[3],[]]
		)

	def testExcessive(self):
		# lowerbounds too high for upperbounds
		self.assertRaises(ValueError,
			Array.from_elements, [1], lowerbounds = (2,), upperbounds = (1,)
		)

	def testNegatives(self):
		a = Array.from_elements([0], lowerbounds = (-1,), upperbounds = (-1,))
		self.assertEqual(a[0], 0)
		self.assertEqual(a[-1], 0)
		# upperbounds at zero
		a = Array.from_elements([1,2], lowerbounds = (-1,), upperbounds = (0,))
		self.assertEqual(a[0], 1)
		self.assertEqual(a[1], 2)
		self.assertEqual(a[-2], 1)
		self.assertEqual(a[-1], 2)

	def testGetElement(self):
		a = Array([1,2,3,4])
		self.assertEqual(a.get_element((0,)), 1)
		self.assertEqual(a.get_element((1,)), 2)
		self.assertEqual(a.get_element((2,)), 3)
		self.assertEqual(a.get_element((3,)), 4)
		self.assertEqual(a.get_element((-1,)), 4)
		self.assertEqual(a.get_element((-2,)), 3)
		self.assertEqual(a.get_element((-3,)), 2)
		self.assertEqual(a.get_element((-4,)), 1)
		self.assertRaises(IndexError, a.get_element, (4,))
		a = Array([[1,2],[3,4]])
		self.assertEqual(a.get_element((0,0)), 1)
		self.assertEqual(a.get_element((0,1,)), 2)
		self.assertEqual(a.get_element((1,0,)), 3)
		self.assertEqual(a.get_element((1,1,)), 4)
		self.assertEqual(a.get_element((-1,-1)), 4)
		self.assertEqual(a.get_element((-1,-2,)), 3)
		self.assertEqual(a.get_element((-2,-1,)), 2)
		self.assertEqual(a.get_element((-2,-2,)), 1)
		self.assertRaises(IndexError, a.get_element, (2,0))
		self.assertRaises(IndexError, a.get_element, (1,2))
		self.assertRaises(IndexError, a.get_element, (0,2))

	def testSQLGetElement(self):
		a = Array([1,2,3,4])
		self.assertEqual(a.sql_get_element((1,)), 1)
		self.assertEqual(a.sql_get_element((2,)), 2)
		self.assertEqual(a.sql_get_element((3,)), 3)
		self.assertEqual(a.sql_get_element((4,)), 4)
		self.assertEqual(a.sql_get_element((0,)), None)
		self.assertEqual(a.sql_get_element((5,)), None)
		self.assertEqual(a.sql_get_element((-1,)), None)
		self.assertEqual(a.sql_get_element((-2,)), None)
		self.assertEqual(a.sql_get_element((-3,)), None)
		self.assertEqual(a.sql_get_element((-4,)), None)
		a = Array([[1,2],[3,4]])
		self.assertEqual(a.sql_get_element((1,1)), 1)
		self.assertEqual(a.sql_get_element((1,2,)), 2)
		self.assertEqual(a.sql_get_element((2,1,)), 3)
		self.assertEqual(a.sql_get_element((2,2,)), 4)
		self.assertEqual(a.sql_get_element((3,1)), None)
		self.assertEqual(a.sql_get_element((1,3)), None)

if __name__ == '__main__':
	from types import ModuleType
	this = ModuleType("this")
	this.__dict__.update(globals())
	unittest.main(this)



##
# .python.decorlib
##
"""
common decorators
"""
import os
import types

def propertydoc(ap):
	"""
	Helper function for extracting an `abstractproperty`'s real documentation.
	"""
	doc = ""
	rstr = ""
	if ap.fget:
		ret = ap.fget.__annotations__.get('return')
		if ret is not None:
			rstr = " -> " + repr(ret)
		if ap.fget.__doc__:
			doc += os.linesep*2 + "GET::" + (os.linesep + ' '*4) + (os.linesep + ' '*4).join(
				[x.strip() for x in ap.fget.__doc__.strip().split(os.linesep)]
			)
	if ap.fset and ap.fset.__doc__:
		doc += os.linesep*2 + "SET::" + (os.linesep + ' '*4) + (os.linesep + ' '*4).join(
			[x.strip() for x in ap.fset.__doc__.strip().split(os.linesep)]
		)
	if ap.fdel and ap.fdel.__doc__:
		doc += os.linesep*2 + "DELETE::" + (os.linesep + ' '*4) + (os.linesep + ' '*4).join(
			[x.strip() for x in ap.fdel.__doc__.strip().split(os.linesep)]
		)
	ap.__doc__ = "<no documentation>" if not doc else (
		"Abstract Property" + rstr + doc
	)
	return ap

class method(object):
	__slots__ = ('callable',)
	def __init__(self, callable):
		self.callable = callable
	def __get__(self, val, typ):
		if val is None:
			return self.callable
		return types.MethodType(self.callable, val)
